from enum import Enum, unique

__all__ = ('RequestMethod', 'ChatType', 'ChatAction', 'ChatMemberStatus',
           'MessageEntityType', 'UpdateType', 'ContentType', 'ParseMode',
           'PollType', 'DiceEmoji', 'InputMediaType', 'InlineQueryResultType',
           'PassportElementType')


@unique
class RequestMethod(str, Enum):
    GET = 'GET'
    POST = 'POST'


@unique
class ChatType(str, Enum):
    PRIVATE = 'private'
    GROUP = 'group'
    SUPERGROUP = 'supergroup'
    CHANNEL = 'channel'


@unique
class ChatAction(str, Enum):
    TYPING = 'typing'
    UPLOAD_PHOTO = 'upload_photo'
    RECORD_VIDEO = 'record_video'
    UPLOAD_VIDEO = 'upload_video'
    RECORD_VOICE = 'record_voice'
    UPLOAD_VOICE = 'upload_voice'
    UPLOAD_DOCUMENT = 'upload_document'
    FIND_LOCATION = 'find_location'
    RECORD_VIDEO_NOTE = 'record_video_note'
    UPLOAD_VIDEO_NOTE = 'upload_video_note'


@unique
class ChatMemberStatus(str, Enum):
    CREATOR = 'creator'
    ADMINISTRATOR = 'administrator'
    MEMBER = 'member'
    RESTRICTED = 'restricted'
    LEFT = 'left'
    KICKED = 'kicked'


@unique
class MessageEntityType(str, Enum):
    MENTION = 'mention'
    HASHTAG = 'hashtag'
    CASHTAG = 'cashtag'
    BOT_COMMAND = 'bot_command'
    URL = 'url'
    EMAIL = 'email'
    PHONE_NUMBER = 'phone_number'
    BOLD = 'bold'
    ITALIC = 'italic'
    CODE = 'code'
    PRE = 'pre'
    TEXT_LINK = 'text_link'
    TEXT_MENTION = 'text_mention'
    UNDERLINE = 'underline'
    STRIKETHROUGH = 'strikethrough'


@unique
class UpdateType(str, Enum):
    MESSAGE = 'message'
    EDITED_MESSAGE = 'edited_message'
    CHANNEL_POST = 'channel_post'
    EDITED_CHANNEL_POST = 'edited_channel_post'
    INLINE_QUERY = 'inline_query'
    CHOSEN_INLINE_RESULT = 'chosen_inline_result'
    CALLBACK_QUERY = 'callback_query'
    SHIPPING_QUERY = 'shipping_query'
    PRE_CHECKOUT_QUERY = 'pre_checkout_query'
    POLL = 'poll'
    POLL_ANSWER = 'poll_answer'
    MY_CHAT_MEMBER = 'my_chat_member'
    CHAT_MEMBER = 'chat_member'


@unique
class ContentType(str, Enum):
    TEXT = 'text'
    AUDIO = 'audio'
    DOCUMENT = 'document'
    ANIMATION = 'animation'
    GAME = 'game'
    PHOTO = 'photo'
    STICKER = 'sticker'
    VIDEO = 'video'
    VIDEO_NOTE = 'video_note'
    VOICE = 'voice'
    CONTACT = 'contact'
    LOCATION = 'location'
    VENUE = 'venue'
    NEW_CHAT_MEMBERS = 'new_chat_members'
    LEFT_CHAT_MEMBER = 'left_chat_member'
    INVOICE = 'invoice'
    SUCCESSFUL_PAYMENT = 'successful_payment'
    CONNECTED_WEBSITE = 'connected_website'
    MIGRATE_TO_CHAT_ID = 'migrate_to_chat_id'
    MIGRATE_FROM_CHAT_ID = 'migrate_from_chat_id'
    PINNED_MESSAGE = 'pinned_message'
    NEW_CHAT_TITLE = 'new_chat_title'
    NEW_CHAT_PHOTO = 'new_chat_photo'
    DELETE_CHAT_PHOTO = 'delete_chat_photo'
    GROUP_CHAT_CREATED = 'group_chat_created'
    PASSPORT_DATA = 'passport_data'


@unique
class ParseMode(str, Enum):
    MARKDOWN = 'Markdown'
    HTML = 'HTML'
    MARKDOWN2 = 'Markdown2'


@unique
class PollType(str, Enum):
    QUIZ = 'quiz'
    REGULAR = 'regular'


@unique
class DiceEmoji(str, Enum):
    DICE = '🎲'
    DARTS = '🎯'
    BASKETBALL = '🏀'
    FOOTBALL = '⚽'
    SLOT_MACHINE = '🎰'
    BOWLING = '🎳'


@unique
class InputMediaType(str, Enum):
    PHOTO = 'photo'
    VIDEO = 'video'
    ANIMATION = 'animation'
    AUDIO = 'audio'
    DOCUMENT = 'document'


@unique
class InlineQueryResultType(str, Enum):
    ARTICLE = 'article'
    PHOTO = 'photo'
    GIF = 'gif'
    MPEG4_GIF = 'mpeg4_gif'
    VIDEO = 'video'
    AUDIO = 'audio'
    VOICE = 'voice'
    DOCUMENT = 'document'
    LOCATION = 'location'
    VENUE = 'venue'
    CONTACT = 'contact'
    GAME = 'game'
    STICKER = 'sticker'


@unique
class PassportElementType(str, Enum):
    PERSONAL_DETAILS = 'personal_details'
    PASSPORT = 'passport'
    DRIVER_LICENSE = 'driver_license'
    IDENTITY_CARD = 'identity_card'
    INTERNAL_PASSPORT = 'internal_passport'
    ADDRESS = 'address'
    UTILITY_BILL = 'utility_bill'
    BANK_STATEMENT = 'bank_statement'
    RENTAL_AGREEMENT = 'rental_agreement'
    PASSPORT_REGISTRATION = 'passport_registration'
    TEMPORARY_REGISTRATION = 'temporary_registration'
    PHONE_NUMBER = 'phone_number'
    EMAIL = 'email'



from typing import Optional

__all__ = ('TelegramError', 'MigrateToChat', 'RetryAfter', 'BadGateway',
           'RestartingTelegram', 'BotBlocked', 'BotKicked')


class TelegramError(Exception):
    pattern: Optional[str] = None

    def __init__(self, error_code: int, description: str) -> None:
        super().__init__(f'{error_code} {description}')
        self.error_code = error_code
        self.description = description

    @classmethod
    def match(cls, description: str) -> bool:
        return cls.pattern is not None and cls.pattern in description.lower()


class MigrateToChat(TelegramError):

    def __init__(self, error_code: int, description: str,
                 chat_id: int) -> None:
        super().__init__(error_code, description)
        self.error_code = error_code
        self.description = description
        self.chat_id = chat_id


class RetryAfter(TelegramError):

    def __init__(self, error_code: int, description: str,
                 retry_after: int) -> None:
        super().__init__(error_code, description)
        self.error_code = error_code
        self.description = description
        self.retry_after = retry_after


class BadGateway(TelegramError):
    pattern = 'bad gateway'


class RestartingTelegram(TelegramError):
    pattern = 'restart'


class BotBlocked(TelegramError):
    pattern = 'bot was blocked by the user'


class BotKicked(TelegramError):
    pattern = 'bot was kicked from a chat'



import asyncio
from enum import Enum
from io import BufferedReader
from pathlib import Path
from typing import (Any, AsyncIterator, Dict, Final, Generator, Iterable, List,
                    Optional, Set, Tuple, Type, TypeVar, Union, cast, get_args,
                    get_origin, get_type_hints)

import attr

from aiotgbot.constants import InputMediaType, ParseMode, PollType

__all__ = ('DataMappingError', 'StreamFile', 'LocalFile', 'BaseTelegram',
           'ResponseParameters', 'APIResponse', 'Update', 'WebhookInfo',
           'User', 'Chat', 'Message', 'MessageId', 'MessageEntity',
           'PhotoSize', 'Audio', 'Document', 'Video', 'Animation', 'Voice',
           'VideoNote', 'Contact', 'Dice', 'Location', 'Venue',
           'ProximityAlertTriggered', 'PollOption', 'PollAnswer', 'Poll',
           'UserProfilePhotos', 'File', 'ReplyMarkup', 'ReplyKeyboardMarkup',
           'KeyboardButton', 'KeyboardButtonPollType', 'ReplyKeyboardRemove',
           'InlineKeyboardMarkup', 'InlineKeyboardButton', 'LoginUrl',
           'CallbackQuery', 'ForceReply', 'ChatPhoto', 'ChatInviteLink',
           'ChatMember', 'ChatMemberUpdated', 'ChatPermissions',
           'ChatLocation', 'BotCommand', 'BotCommandScopeDefault',
           'BotCommandScopeAllPrivateChats', 'BotCommandScopeAllGroupChats',
           'BotCommandScopeChatAdministrators', 'BotCommandScopeChat',
           'BotCommandScopeAllChatAdministrators', 'BotCommandScopeChatMember',
           'InputFile', 'BotCommandScope', 'InputMedia', 'InputMediaPhoto',
           'InputMediaVideo', 'InputMediaAnimation', 'InputMediaAudio',
           'InputMediaDocument', 'Sticker', 'StickerSet', 'MaskPosition',
           'InlineQuery', 'InlineQueryResult', 'InlineQueryResultArticle',
           'InlineQueryResultPhoto', 'InlineQueryResultGif',
           'InlineQueryResultMpeg4Gif', 'InlineQueryResultVideo',
           'InlineQueryResultAudio', 'InlineQueryResultVoice',
           'InlineQueryResultDocument', 'InlineQueryResultLocation',
           'InlineQueryResultVenue', 'InlineQueryResultContact',
           'InlineQueryResultGame', 'InlineQueryResultCachedPhoto',
           'InlineQueryResultCachedGif', 'InlineQueryResultCachedMpeg4Gif',
           'InlineQueryResultCachedSticker', 'InlineQueryResultCachedDocument',
           'InlineQueryResultCachedVideo', 'InlineQueryResultCachedVoice',
           'InlineQueryResultCachedAudio', 'InputMessageContent',
           'InputTextMessageContent', 'InputLocationMessageContent',
           'InputVenueMessageContent', 'InputContactMessageContent',
           'ChosenInlineResult', 'LabeledPrice', 'Invoice', 'ShippingAddress',
           'OrderInfo', 'ShippingOption', 'SuccessfulPayment', 'ShippingQuery',
           'PreCheckoutQuery', 'PassportData', 'PassportFile',
           'EncryptedPassportElement', 'EncryptedCredentials',
           'PassportElementError', 'PassportElementErrorDataField',
           'PassportElementErrorFrontSide', 'PassportElementErrorReverseSide',
           'PassportElementErrorSelfie', 'PassportElementErrorFile',
           'PassportElementErrorFiles', 'PassportElementErrorTranslationFile',
           'PassportElementErrorTranslationFiles',
           'PassportElementErrorUnspecified', 'Game', 'CallbackGame',
           'GameHighScore')


class DataMappingError(BaseException):
    pass


@attr.s(frozen=True, auto_attribs=True)
class StreamFile:
    content: AsyncIterator[bytes]
    name: str
    content_type: Optional[str] = None


class LocalFile:

    def __init__(
        self, path: Union[str, Path], content_type: Optional[str] = None,
    ) -> None:
        self._path: Final[Path] = (path if isinstance(path, Path)
                                   else Path(path))
        self._content_type: Final[Optional[str]] = content_type

    @property
    def name(self) -> str:
        return self._path.name

    @property
    def content_type(self) -> Optional[str]:
        return self._content_type

    @property
    async def content(self) -> AsyncIterator[bytes]:
        loop = asyncio.get_running_loop()
        reader = cast(BufferedReader, await loop.run_in_executor(
            None, self._path.open, 'rb'))
        try:
            chunk = await loop.run_in_executor(None, reader.read, 2 ** 16)
            while len(chunk) > 0:
                yield chunk
                chunk = await loop.run_in_executor(None, reader.read, 2 ** 16)
        finally:
            await loop.run_in_executor(None, reader.close)


def _is_tuple(_type: Any) -> bool:
    return get_origin(_type) is tuple


def _is_list(_type: Any) -> bool:
    return get_origin(_type) is list


def _is_union(_type: Any) -> bool:
    return get_origin(_type) is Union


def _is_optional(_type: Any) -> bool:
    return _type is Any or (_is_union(_type) and type(None) in get_args(_type))


def _is_attr_union(_type: Any) -> bool:
    return _is_union(_type) and all(attr.has(arg_type) or arg_type is _NoneType
                                    for arg_type in get_args(_type))


_NoneType: Type[None] = type(None)
_FieldType = Union[int, str, bool, float, Tuple[Any, ...], List[Any],
                   Dict[str, Any], 'BaseTelegram']
_HintsGenerator = Generator[Tuple[str, str, Any], None, None]


@attr.s(frozen=True)
class BaseTelegram:

    def to_dict(self) -> Dict[str, Any]:
        _dict: Dict[str, Any] = {}
        for _attr in attr.fields(type(self)):
            value = getattr(self, _attr.name)
            key = _attr.name.rstrip('_')
            if isinstance(value, BaseTelegram):
                _dict[key] = value.to_dict()
            elif isinstance(value, (tuple, list)):
                _dict[key] = BaseTelegram._to_list(value)
            elif isinstance(value, (int, str, bool, float)):
                _dict[key] = value
            elif isinstance(value, Enum):
                _dict[key] = value.value
            elif value is None:
                continue
            else:
                raise TypeError(f'"{value}" has unsupported type')

        return _dict

    @classmethod
    def from_dict(cls: Type['_Telegram'], data: Dict[str, Any]) -> '_Telegram':
        return cast(_Telegram, BaseTelegram._handle_object(cls, data))

    @staticmethod
    def _to_list(value: Iterable[Any]) -> List[Any]:
        _list: List[Any] = []
        for item in value:
            if isinstance(item, (int, str, bool, float)):
                _list.append(item)
            elif isinstance(item, (tuple, list)):
                _list.append(BaseTelegram._to_list(item))
            elif isinstance(item, BaseTelegram):
                _list.append(item.to_dict())
            elif isinstance(value, Enum):
                _list.append(value.value)
            else:
                raise TypeError(f'"{item}" has unsupported type')

        return _list

    @staticmethod
    def _get_type_hints(_type: Any) -> Tuple[Tuple[str, str, Any], ...]:
        return tuple((field.rstrip('_'), field, _type)
                     for field, _type in get_type_hints(_type).items())

    @staticmethod
    def _handle_object(_type: Any, data: Dict[str, Any]) -> Any:
        assert issubclass(_type, BaseTelegram)
        assert attr.has(_type)
        type_hints = _type._get_type_hints(_type)
        required = frozenset(field for field, _, _type in type_hints
                             if not _is_optional(_type))
        filled = frozenset(key for key, value in data.items()
                           if value is not None)
        if not required <= filled:
            keys = ', '.join(required - filled)
            raise DataMappingError(f'Data without required keys: {keys}')
        params = {field: BaseTelegram._handle_field(_type, data.get(key))
                  for key, field, _type in type_hints}

        return cast(Any, _type)(**params)

    @staticmethod
    def _handle_field(_type: Any, value: Any) -> Optional[_FieldType]:
        if _type in (int, str, bool, float) and isinstance(value, _type):
            return cast(_FieldType, value)
        elif _type in (int, str, bool, float):
            message = f'"{value}" is not instance of type "{_type.__name__}"'
            raise DataMappingError(message)
        elif _type is _NoneType and value is None:
            return None
        elif _type is _NoneType:
            raise DataMappingError(f'"{value}" is not None')
        elif _type is Any:
            return cast(_FieldType, value)
        elif _is_tuple(_type) and isinstance(value, list):
            return tuple(BaseTelegram._handle_field(get_args(_type)[0], item)
                         for item in value)
        elif _is_list(_type) and isinstance(value, list):
            return [BaseTelegram._handle_field(get_args(_type)[0], item)
                    for item in value]
        elif _is_list(_type) or _is_tuple(_type):
            raise DataMappingError(f'Data "{value}" is not list')
        elif _is_optional(_type) and value is None:
            return None
        elif _is_optional(_type) and len(get_args(_type)) == 2:
            return BaseTelegram._handle_field(get_args(_type)[0], value)
        elif _is_attr_union(_type) and isinstance(value, dict):
            types: List[Tuple[int, Any]] = []
            for arg_type in get_args(_type):
                fields: Set[str] = set(key.rstrip('_') for key
                                       in get_type_hints(arg_type).keys())
                data_keys: Set[str] = set(value.keys())
                if not data_keys <= fields:
                    continue
                types.append((len(fields & data_keys), arg_type))
            if len(types) == 0:
                arg_types = ', '.join(t.__name__ for t in get_args(_type))
                message = f'Data "{value}" not match any of "{arg_types}"'
                raise DataMappingError(message)
            types = sorted(types, key=lambda t: t[0], reverse=True)
            return BaseTelegram._handle_field(types[0][1], value)
        elif attr.has(_type) and isinstance(value, dict):
            return cast(
                BaseTelegram,
                BaseTelegram._handle_object(_type, value)
            )
        else:
            message = f'Data "{value}" not match field type "{_type}"'
            raise DataMappingError(message)


_Telegram = TypeVar('_Telegram', bound=BaseTelegram)


@attr.s(auto_attribs=True)
class ResponseParameters(BaseTelegram):
    migrate_to_chat_id: Optional[int] = None
    retry_after: Optional[int] = None


@attr.s(auto_attribs=True)
class APIResponse(BaseTelegram):
    ok: bool
    result: Any
    error_code: Optional[int] = None
    description: Optional[str] = None
    parameters: Optional[ResponseParameters] = None


@attr.s(auto_attribs=True)
class Update(BaseTelegram):
    update_id: int
    message: Optional['Message'] = None
    edited_message: Optional['Message'] = None
    channel_post: Optional['Message'] = None
    edited_channel_post: Optional['Message'] = None
    inline_query: Optional['InlineQuery'] = None
    chosen_inline_result: Optional['ChosenInlineResult'] = None
    callback_query: Optional['CallbackQuery'] = None
    shipping_query: Optional['ShippingQuery'] = None
    pre_checkout_query: Optional['PreCheckoutQuery'] = None
    poll: Optional['Poll'] = None
    poll_answer: Optional['PollAnswer'] = None
    my_chat_member: Optional['ChatMemberUpdated'] = None
    chat_member: Optional['ChatMemberUpdated'] = None


@attr.s(auto_attribs=True)
class WebhookInfo(BaseTelegram):
    allowed_updates: Tuple[str, ...]
    url: Optional[str] = None
    has_custom_certificate: Optional[bool] = None
    pending_update_count: Optional[int] = None
    ip_address: Optional[str] = None
    last_error_date: Optional[int] = None
    last_error_message: Optional[str] = None
    max_connections: Optional[int] = None


@attr.s(auto_attribs=True)
class User(BaseTelegram):
    id: int
    is_bot: bool
    first_name: str
    last_name: Optional[str] = None
    username: Optional[str] = None
    language_code: Optional[str] = None
    can_join_groups: Optional[bool] = None
    can_read_all_group_messages: Optional[bool] = None
    supports_inline_queries: Optional[bool] = None


@attr.s(auto_attribs=True)
class Chat(BaseTelegram):
    id: int
    type: str
    title: Optional[str] = None
    username: Optional[str] = None
    first_name: Optional[str] = None
    last_name: Optional[str] = None
    photo: Optional['ChatPhoto'] = None
    bio: Optional[str] = None
    description: Optional[str] = None
    invite_link: Optional[str] = None
    pinned_message: Optional['Message'] = None
    permissions: Optional['ChatPermissions'] = None
    slow_mode_delay: Optional[int] = None
    sticker_set_name: Optional[str] = None
    can_set_sticker_set: Optional[bool] = None
    linked_chat_id: Optional[int] = None
    location: Optional['ChatLocation'] = None


@attr.s(auto_attribs=True)
class Message(BaseTelegram):
    message_id: int
    date: int
    chat: Chat
    from_: Optional[User] = None
    sender_chat: Optional[Chat] = None
    forward_from: Optional[User] = None
    forward_from_chat: Optional[Chat] = None
    forward_from_message_id: Optional[int] = None
    forward_signature: Optional[str] = None
    forward_sender_name: Optional[str] = None
    forward_date: Optional[int] = None
    reply_to_message: Optional['Message'] = None
    via_bot: Optional[User] = None
    edit_date: Optional[int] = None
    media_group_id: Optional[str] = None
    author_signature: Optional[str] = None
    text: Optional[str] = None
    entities: Optional[Tuple['MessageEntity', ...]] = None
    caption_entities: Optional[Tuple['MessageEntity', ...]] = None
    audio: Optional['Audio'] = None
    document: Optional['Document'] = None
    animation: Optional['Animation'] = None
    game: Optional['Game'] = None
    photo: Optional[Tuple['PhotoSize', ...]] = None
    sticker: Optional['Sticker'] = None
    video: Optional['Video'] = None
    voice: Optional['Voice'] = None
    video_note: Optional['VideoNote'] = None
    caption: Optional[str] = None
    contact: Optional['Contact'] = None
    dice: Optional['Dice'] = None
    location: Optional['Location'] = None
    venue: Optional['Venue'] = None
    poll: Optional['Poll'] = None
    new_chat_members: Optional[Tuple[User, ...]] = None
    left_chat_member: Optional[User] = None
    new_chat_title: Optional[str] = None
    new_chat_photo: Optional[Tuple['PhotoSize', ...]] = None
    delete_chat_photo: Optional[bool] = None
    group_chat_created: Optional[bool] = None
    supergroup_chat_created: Optional[bool] = None
    channel_chat_created: Optional[bool] = None
    message_auto_delete_timer_changed: Optional['MADTC'] = None
    migrate_to_chat_id: Optional[int] = None
    migrate_from_chat_id: Optional[int] = None
    pinned_message: Optional['Message'] = None
    invoice: Optional['Invoice'] = None
    successful_payment: Optional['SuccessfulPayment'] = None
    connected_website: Optional[str] = None
    passport_data: Optional['PassportData'] = None
    proximity_alert_triggered: Optional['ProximityAlertTriggered'] = None
    voice_chat_scheduled: Optional['VoiceChatScheduled'] = None
    voice_chat_started: Optional['VoiceChatStarted'] = None
    voice_chat_ended: Optional['VoiceChatEnded'] = None
    voice_chat_participants_invited: Optional['VCPI'] = None
    reply_markup: Optional['InlineKeyboardMarkup'] = None


@attr.s(auto_attribs=True)
class MessageId(BaseTelegram):
    message_id: int


@attr.s(auto_attribs=True)
class MessageEntity(BaseTelegram):
    type: str
    offset: int
    length: int
    url: Optional[str] = None
    user: Optional[User] = None
    language: Optional[str] = None


@attr.s(auto_attribs=True)
class PhotoSize(BaseTelegram):
    file_id: str
    file_unique_id: str
    width: int
    height: int
    file_size: int


@attr.s(auto_attribs=True)
class Audio(BaseTelegram):
    file_id: str
    file_unique_id: str
    duration: int
    performer: Optional[str] = None
    title: Optional[str] = None
    file_name: Optional[str] = None
    mime_type: Optional[str] = None
    file_size: Optional[int] = None
    thumb: Optional[PhotoSize] = None


@attr.s(auto_attribs=True)
class Document(BaseTelegram):
    file_id: str
    file_unique_id: str
    thumb: Optional[PhotoSize] = None
    file_name: Optional[str] = None
    mime_type: Optional[str] = None
    file_size: Optional[int] = None


@attr.s(auto_attribs=True)
class Video(BaseTelegram):
    file_id: str
    file_unique_id: str
    width: int
    height: int
    duration: int
    thumb: Optional[PhotoSize] = None
    file_name: Optional[str] = None
    mime_type: Optional[str] = None
    file_size: Optional[int] = None


@attr.s(auto_attribs=True)
class Animation(BaseTelegram):
    file_id: str
    file_unique_id: str
    thumb: Optional[PhotoSize] = None
    file_name: Optional[str] = None
    mime_type: Optional[str] = None
    file_size: Optional[int] = None


@attr.s(auto_attribs=True)
class Voice(BaseTelegram):
    file_id: str
    file_unique_id: str
    duration: int
    mime_type: Optional[str] = None
    file_size: Optional[int] = None


@attr.s(auto_attribs=True)
class VideoNote(BaseTelegram):
    file_id: str
    file_unique_id: str
    length: int
    duration: int
    thumb: Optional[PhotoSize] = None
    file_size: Optional[int] = None


@attr.s(auto_attribs=True)
class Contact(BaseTelegram):
    phone_number: str
    first_name: str
    last_name: Optional[str] = None
    user_id: Optional[int] = None
    vcard: Optional[int] = None


@attr.s(auto_attribs=True)
class Dice(BaseTelegram):
    emoji: str
    value: int


@attr.s(auto_attribs=True)
class Location(BaseTelegram):
    longitude: float
    latitude: float
    horizontal_accuracy: Optional[float] = None
    live_period: Optional[int] = None
    heading: Optional[int] = None
    proximity_alert_radius: Optional[int] = None


@attr.s(auto_attribs=True)
class Venue(BaseTelegram):
    location: Location
    title: str
    address: str
    foursquare_id: Optional[str] = None
    foursquare_type: Optional[str] = None
    google_place_id: Optional[str] = None
    google_place_type: Optional[str] = None


@attr.s(auto_attribs=True)
class VoiceChatStarted(BaseTelegram):
    pass


@attr.s(auto_attribs=True)
class VoiceChatEnded(BaseTelegram):
    duration: int


@attr.s(auto_attribs=True)
class VoiceChatParticipantsInvited(BaseTelegram):
    users: Optional[Tuple[User, ...]] = None


VCPI = VoiceChatParticipantsInvited


@attr.s(auto_attribs=True)
class ProximityAlertTriggered(BaseTelegram):
    traveler: User
    watcher: User
    distance: int


@attr.s(auto_attribs=True)
class MessageAutoDeleteTimerChanged(BaseTelegram):
    message_auto_delete_time: int


MADTC = MessageAutoDeleteTimerChanged


@attr.s(auto_attribs=True)
class VoiceChatScheduled(BaseTelegram):
    start_date: int


@attr.s(auto_attribs=True)
class PollOption(BaseTelegram):
    text: str
    voter_count: int


@attr.s(auto_attribs=True)
class PollAnswer(BaseTelegram):
    poll_id: str
    user: User
    option_ids: Tuple[int, ...]


@attr.s(auto_attribs=True)
class Poll(BaseTelegram):
    id: str
    question: str
    options: Tuple[PollOption, ...]
    total_voter_count: int
    is_closed: bool
    is_anonymous: bool
    type_: str
    allows_multiple_answers: bool
    correct_option_id: Optional[int] = None
    explanation: Optional[str] = None
    explanation_entities: Optional[Tuple[MessageEntity, ...]] = None
    open_period: Optional[int] = None
    close_date: Optional[int] = None


@attr.s(auto_attribs=True)
class UserProfilePhotos(BaseTelegram):
    total_count: int
    photos: Tuple[Tuple[PhotoSize, ...], ...]


@attr.s(auto_attribs=True)
class File(BaseTelegram):
    file_id: str
    file_unique_id: str
    file_size: Optional[int] = None
    file_path: Optional[str] = None


ReplyMarkup = Union['InlineKeyboardMarkup',
                    'ReplyKeyboardMarkup',
                    'ReplyKeyboardRemove',
                    'ForceReply']


@attr.s(auto_attribs=True)
class ReplyKeyboardMarkup(BaseTelegram):
    keyboard: List[List['KeyboardButton']]
    resize_keyboard: Optional[bool] = None
    one_time_keyboard: Optional[bool] = None
    input_field_placeholder: Optional[str] = None
    selective: Optional[bool] = None


@attr.s(auto_attribs=True)
class KeyboardButton(BaseTelegram):
    text: str
    request_contact: Optional[bool] = None
    request_location: Optional[bool] = None
    request_poll: Optional['KeyboardButtonPollType'] = None


@attr.s(auto_attribs=True)
class KeyboardButtonPollType(BaseTelegram):
    type_: PollType


@attr.s(auto_attribs=True)
class ReplyKeyboardRemove(BaseTelegram):
    remove_keyboard: bool
    selective: Optional[bool] = None


@attr.s(auto_attribs=True)
class InlineKeyboardMarkup(BaseTelegram):
    inline_keyboard: List[List['InlineKeyboardButton']]


@attr.s(auto_attribs=True)
class InlineKeyboardButton(BaseTelegram):
    text: str
    url: Optional[str] = None
    login_url: Optional['LoginUrl'] = None
    callback_data: Optional[str] = None
    switch_inline_query: Optional[str] = None
    switch_inline_query_current_chat: Optional[str] = None
    callback_game: Optional['CallbackGame'] = None
    pay: Optional[bool] = None


@attr.s(auto_attribs=True)
class LoginUrl(BaseTelegram):
    url: str
    forward_text: Optional[str] = None
    bot_username: Optional[str] = None
    request_write_access: Optional[bool] = None


@attr.s(auto_attribs=True)
class CallbackQuery(BaseTelegram):
    id: str
    from_: User
    chat_instance: str
    message: Optional[Message] = None
    inline_message_id: Optional[str] = None
    data: Optional[str] = None
    game_short_name: Optional[str] = None


@attr.s(auto_attribs=True)
class ForceReply(BaseTelegram):
    force_reply: bool
    input_field_placeholder: Optional[str] = None
    selective: Optional[bool] = None


@attr.s(auto_attribs=True)
class ChatPhoto(BaseTelegram):
    small_file_id: str
    small_file_unique_id: str
    big_file_id: str
    big_file_unique_id: str


@attr.s(auto_attribs=True)
class ChatInviteLink(BaseTelegram):
    invite_link: str
    creator: User
    is_primary: bool
    is_revoked: bool
    expire_date: Optional[int] = None
    member_limit: Optional[int] = None


@attr.s(auto_attribs=True)
class ChatMember(BaseTelegram):
    user: User
    status: str
    custom_title: Optional[str] = None
    is_anonymous: Optional[bool] = None
    until_date: Optional[int] = None
    can_be_edited: Optional[bool] = None
    can_manage_chat: Optional[bool] = None
    can_change_info: Optional[bool] = None
    can_post_messages: Optional[bool] = None
    can_edit_messages: Optional[bool] = None
    can_delete_messages: Optional[bool] = None
    can_manage_voice_chats: Optional[bool] = None
    can_invite_users: Optional[bool] = None
    can_restrict_members: Optional[bool] = None
    can_pin_messages: Optional[bool] = None
    can_promote_members: Optional[bool] = None
    is_member: Optional[bool] = None
    can_send_messages: Optional[bool] = None
    can_send_media_messages: Optional[bool] = None
    can_send_other_messages: Optional[bool] = None
    can_add_web_page_previews: Optional[bool] = None
    can_send_polls: Optional[bool] = None


@attr.s(auto_attribs=True)
class ChatMemberUpdated(BaseTelegram):
    chat: Chat
    from_: User
    date: int
    old_chat_member: ChatMember
    new_chat_member: ChatMember
    invite_link: Optional[ChatInviteLink] = None


@attr.s(auto_attribs=True)
class ChatPermissions(BaseTelegram):
    can_send_messages: Optional[bool] = None
    can_send_media_messages: Optional[bool] = None
    can_send_polls: Optional[bool] = None
    can_send_other_messages: Optional[bool] = None
    can_add_web_page_previews: Optional[bool] = None
    can_change_info: Optional[bool] = None
    can_invite_users: Optional[bool] = None
    can_pin_messages: Optional[bool] = None


@attr.s(auto_attribs=True)
class ChatLocation(BaseTelegram):
    location: Location
    address: str


@attr.s(auto_attribs=True)
class BotCommand(BaseTelegram):
    command: str
    description: str


@attr.s(auto_attribs=True)
class BotCommandScopeDefault(BaseTelegram):
    type: str = 'default'


@attr.s(auto_attribs=True)
class BotCommandScopeAllPrivateChats(BaseTelegram):
    type: str = 'all_private_chats'


@attr.s(auto_attribs=True)
class BotCommandScopeAllGroupChats(BaseTelegram):
    type: str = 'all_group_chats'


@attr.s(auto_attribs=True)
class BotCommandScopeAllChatAdministrators(BaseTelegram):
    type: str = 'all_chat_administrators'


@attr.s(auto_attribs=True)
class BotCommandScopeChat(BaseTelegram):
    chat_id: Union[int, str]
    type: str = 'chat'


@attr.s(auto_attribs=True)
class BotCommandScopeChatAdministrators(BaseTelegram):
    chat_id: Union[int, str]
    type: str = 'chat'


@attr.s(auto_attribs=True)
class BotCommandScopeChatMember(BaseTelegram):
    chat_id: Union[int, str]
    user_id: int
    type: str = 'chat'


BotCommandScope = Union[BotCommandScopeDefault,
                        BotCommandScopeAllPrivateChats,
                        BotCommandScopeAllGroupChats,
                        BotCommandScopeAllChatAdministrators,
                        BotCommandScopeChat,
                        BotCommandScopeChatAdministrators,
                        BotCommandScopeChatMember]

InputFile = Union[LocalFile, StreamFile]


@attr.s(auto_attribs=True)
class InputMedia(BaseTelegram):
    media: Union[str, InputFile]
    caption: Optional[str] = None
    parse_mode: Optional[str] = None
    caption_entities: Optional[Iterable[MessageEntity]] = None

    def to_dict(self) -> Dict[str, Any]:
        if not isinstance(self.media, str):
            raise TypeError('To serialize this object, the media attribute '
                            'type must be a string')
        return super().to_dict()


@attr.s(auto_attribs=True)
class InputMediaPhoto(InputMedia):
    type: str = attr.ib(default=InputMediaType.PHOTO, init=False)


@attr.s(auto_attribs=True)
class InputMediaVideo(InputMedia):
    type: str = attr.ib(default=InputMediaType.VIDEO, init=False)
    thumb: Optional[str] = None
    width: Optional[int] = None
    height: Optional[int] = None
    duration: Optional[int] = None
    supports_streaming: Optional[bool] = None


@attr.s(auto_attribs=True)
class InputMediaAnimation(InputMedia):
    type: str = attr.ib(default=InputMediaType.ANIMATION, init=False)
    thumb: Optional[str] = None
    width: Optional[int] = None
    height: Optional[int] = None
    duration: Optional[int] = None


@attr.s(auto_attribs=True)
class InputMediaAudio(InputMedia):
    type: str = attr.ib(default=InputMediaType.AUDIO, init=False)
    thumb: Optional[str] = None
    duration: Optional[int] = None
    performer: Optional[str] = None
    title: Optional[str] = None


@attr.s(auto_attribs=True)
class InputMediaDocument(InputMedia):
    type: str = attr.ib(default=InputMediaType.DOCUMENT, init=False)
    thumb: Optional[str] = None
    disable_content_type_detection: Optional[bool] = None


@attr.s(auto_attribs=True)
class Sticker(BaseTelegram):
    file_id: str
    file_unique_id: str
    width: int
    height: int
    is_animated: bool
    thumb: Optional[PhotoSize] = None
    emoji: Optional[str] = None
    set_name: Optional[str] = None
    mask_position: Optional['MaskPosition'] = None
    file_size: Optional[int] = None


@attr.s(auto_attribs=True)
class StickerSet(BaseTelegram):
    name: str
    title: str
    is_animated: bool
    contains_masks: bool
    stickers: Tuple[Sticker, ...]
    thumb: Optional[PhotoSize] = None


@attr.s(auto_attribs=True)
class MaskPosition(BaseTelegram):
    point: str
    x_shift: float
    y_shift: float
    scale: float


@attr.s(auto_attribs=True)
class InlineQuery(BaseTelegram):
    id: str
    from_: User
    query: str
    offset: str
    chat_type: Optional[str] = None
    location: Optional[Location] = None


InlineQueryResult = Union['InlineQueryResultCachedAudio',
                          'InlineQueryResultCachedDocument',
                          'InlineQueryResultCachedGif',
                          'InlineQueryResultCachedMpeg4Gif',
                          'InlineQueryResultCachedPhoto',
                          'InlineQueryResultCachedSticker',
                          'InlineQueryResultCachedVideo',
                          'InlineQueryResultCachedVoice',
                          'InlineQueryResultArticle',
                          'InlineQueryResultAudio',
                          'InlineQueryResultContact',
                          'InlineQueryResultGame',
                          'InlineQueryResultDocument',
                          'InlineQueryResultGif',
                          'InlineQueryResultLocation',
                          'InlineQueryResultMpeg4Gif',
                          'InlineQueryResultPhoto',
                          'InlineQueryResultVenue',
                          'InlineQueryResultVideo',
                          'InlineQueryResultVoice']


@attr.s(auto_attribs=True)
class InlineQueryResultArticle(BaseTelegram):
    type: str
    id: str
    title: str
    input_message_content: 'InputMessageContent'
    reply_markup: Optional[InlineKeyboardMarkup] = None
    url: Optional[str] = None
    hide_url: Optional[bool] = None
    description: Optional[str] = None
    thumb_url: Optional[str] = None
    thumb_width: Optional[int] = None
    thumb_height: Optional[int] = None


@attr.s(auto_attribs=True)
class InlineQueryResultPhoto(BaseTelegram):
    type: str
    td: str
    photo_url: str
    thumb_url: str
    photo_width: Optional[int] = None
    photo_height: Optional[int] = None
    title: Optional[str] = None
    description: Optional[str] = None
    caption: Optional[str] = None
    parse_mode: Optional[ParseMode] = None
    caption_entities: Optional[Iterable[MessageEntity]] = None
    reply_markup: Optional[InlineKeyboardMarkup] = None
    input_message_content: Optional['InputMessageContent'] = None


@attr.s(auto_attribs=True)
class InlineQueryResultGif(BaseTelegram):
    type: str
    id: str
    gif_url: str
    thumb_url: str
    gif_width: Optional[int] = None
    gif_height: Optional[int] = None
    title: Optional[str] = None
    caption: Optional[str] = None
    parse_mode: Optional[ParseMode] = None
    caption_entities: Optional[Iterable[MessageEntity]] = None
    reply_markup: Optional[InlineKeyboardMarkup] = None
    input_message_content: Optional['InputMessageContent'] = None


@attr.s(auto_attribs=True)
class InlineQueryResultMpeg4Gif(BaseTelegram):
    type: str
    id: str
    mpeg4_url: str
    thumb_url: str
    mpeg4_width: Optional[int] = None
    mpeg4_height: Optional[int] = None
    title: Optional[str] = None
    caption: Optional[str] = None
    parse_mode: Optional[ParseMode] = None
    caption_entities: Optional[Iterable[MessageEntity]] = None
    reply_markup: Optional[InlineKeyboardMarkup] = None
    input_message_content: Optional['InputMessageContent'] = None


@attr.s(auto_attribs=True)
class InlineQueryResultVideo(BaseTelegram):
    type: str
    id: str
    video_url: str
    mime_type: str
    thumb_url: str
    title: str
    caption: Optional[str] = None
    parse_mode: Optional[ParseMode] = None
    caption_entities: Optional[Iterable[MessageEntity]] = None
    video_width: Optional[int] = None
    video_height: Optional[int] = None
    video_duration: Optional[int] = None
    description: Optional[str] = None
    reply_markup: Optional[InlineKeyboardMarkup] = None
    input_message_content: Optional['InputMessageContent'] = None


@attr.s(auto_attribs=True)
class InlineQueryResultAudio(BaseTelegram):
    type: str
    id: str
    audio_url: str
    title: str
    caption: Optional[str] = None
    parse_mode: Optional[ParseMode] = None
    caption_entities: Optional[Iterable[MessageEntity]] = None
    performer: Optional[str] = None
    audio_duration: Optional[int] = None
    reply_markup: Optional[InlineKeyboardMarkup] = None
    input_message_content: Optional['InputMessageContent'] = None


@attr.s(auto_attribs=True)
class InlineQueryResultVoice(BaseTelegram):
    type: str
    id: str
    voice_url: str
    title: str
    caption: Optional[str] = None
    parse_mode: Optional[ParseMode] = None
    caption_entities: Optional[Iterable[MessageEntity]] = None
    voice_duration: Optional[int] = None
    reply_markup: Optional[InlineKeyboardMarkup] = None
    input_message_content: Optional['InputMessageContent'] = None


@attr.s(auto_attribs=True)
class InlineQueryResultDocument(BaseTelegram):
    type: str
    id: str
    title: str
    document_url: str
    mime_type: str
    caption: Optional[str] = None
    parse_mode: Optional[ParseMode] = None
    caption_entities: Optional[Iterable[MessageEntity]] = None
    description: Optional[str] = None
    reply_markup: Optional[InlineKeyboardMarkup] = None
    input_message_content: Optional['InputMessageContent'] = None
    thumb_url: Optional[str] = None
    thumb_width: Optional[int] = None
    thumb_height: Optional[int] = None


@attr.s(auto_attribs=True)
class InlineQueryResultLocation(BaseTelegram):
    type: str
    id: str
    latitude: float
    longitude: float
    title: str
    horizontal_accuracy: Optional[float] = None
    live_period: Optional[int] = None
    heading: Optional[int] = None
    proximity_alert_radius: Optional[int] = None
    reply_markup: Optional[InlineKeyboardMarkup] = None
    input_message_content: Optional['InputMessageContent'] = None
    thumb_url: Optional[str] = None
    thumb_width: Optional[int] = None
    thumb_height: Optional[int] = None


@attr.s(auto_attribs=True)
class InlineQueryResultVenue(BaseTelegram):
    type: str
    id: str
    latitude: float
    longitude: float
    title: str
    address: str
    foursquare_id: Optional[str] = None
    foursquare_type: Optional[str] = None
    google_place_id: Optional[str] = None
    google_place_type: Optional[str] = None
    reply_markup: Optional[InlineKeyboardMarkup] = None
    input_message_content: Optional['InputMessageContent'] = None
    thumb_url: Optional[str] = None
    thumb_width: Optional[int] = None
    thumb_height: Optional[int] = None


@attr.s(auto_attribs=True)
class InlineQueryResultContact(BaseTelegram):
    type: str
    id: str
    phone_number: str
    first_name: str
    last_name: Optional[str] = None
    vcard: Optional[str] = None
    reply_markup: Optional[InlineKeyboardMarkup] = None
    input_message_content: Optional['InputMessageContent'] = None
    thumb_url: Optional[str] = None
    thumb_width: Optional[int] = None
    thumb_height: Optional[int] = None


@attr.s(auto_attribs=True)
class InlineQueryResultGame(BaseTelegram):
    type: str
    id: str
    game_short_name: str
    reply_markup: Optional[InlineKeyboardMarkup] = None


@attr.s(auto_attribs=True)
class InlineQueryResultCachedPhoto(BaseTelegram):
    type: str
    id: str
    photofileid: str
    title: Optional[str] = None
    description: Optional[str] = None
    caption: Optional[str] = None
    parse_mode: Optional[ParseMode] = None
    caption_entities: Optional[Iterable[MessageEntity]] = None
    reply_markup: Optional[InlineKeyboardMarkup] = None
    input_message_content: Optional['InputMessageContent'] = None


@attr.s(auto_attribs=True)
class InlineQueryResultCachedGif(BaseTelegram):
    type: str
    id: str
    gif_file_id: str
    title: Optional[str] = None
    caption: Optional[str] = None
    parse_mode: Optional[ParseMode] = None
    caption_entities: Optional[Iterable[MessageEntity]] = None
    reply_markup: Optional[InlineKeyboardMarkup] = None
    input_message_content: Optional['InputMessageContent'] = None


@attr.s(auto_attribs=True)
class InlineQueryResultCachedMpeg4Gif(BaseTelegram):
    type: str
    id: str
    mpeg4_file_id: str
    title: Optional[str] = None
    caption: Optional[str] = None
    parse_mode: Optional[ParseMode] = None
    caption_entities: Optional[Iterable[MessageEntity]] = None
    reply_markup: Optional[InlineKeyboardMarkup] = None
    input_message_content: Optional['InputMessageContent'] = None


@attr.s(auto_attribs=True)
class InlineQueryResultCachedSticker(BaseTelegram):
    type: str
    id: str
    sticker_file_id: str
    reply_markup: Optional[InlineKeyboardMarkup] = None
    input_message_content: Optional['InputMessageContent'] = None


@attr.s(auto_attribs=True)
class InlineQueryResultCachedDocument(BaseTelegram):
    type: str
    id: str
    title: str
    document_file_id: str
    description: Optional[str] = None
    caption: Optional[str] = None
    parse_mode: Optional[ParseMode] = None
    caption_entities: Optional[Iterable[MessageEntity]] = None
    reply_markup: Optional[InlineKeyboardMarkup] = None
    input_message_content: Optional['InputMessageContent'] = None


@attr.s(auto_attribs=True)
class InlineQueryResultCachedVideo(BaseTelegram):
    type: str
    id: str
    video_file_id: str
    title: str
    description: Optional[str] = None
    caption: Optional[str] = None
    parse_mode: Optional[ParseMode] = None
    caption_entities: Optional[Iterable[MessageEntity]] = None
    reply_markup: Optional[InlineKeyboardMarkup] = None
    input_message_content: Optional['InputMessageContent'] = None


@attr.s(auto_attribs=True)
class InlineQueryResultCachedVoice(BaseTelegram):
    type: str
    id: str
    voice_file_id: str
    title: str
    caption: Optional[str] = None
    parse_mode: Optional[ParseMode] = None
    caption_entities: Optional[Iterable[MessageEntity]] = None
    reply_markup: Optional[InlineKeyboardMarkup] = None
    input_message_content: Optional['InputMessageContent'] = None


@attr.s(auto_attribs=True)
class InlineQueryResultCachedAudio(BaseTelegram):
    type: str
    id: str
    audio_file_id: str
    caption: Optional[str] = None
    parse_mode: Optional[ParseMode] = None
    caption_entities: Optional[Iterable[MessageEntity]] = None
    reply_markup: Optional[InlineKeyboardMarkup] = None
    input_message_content: Optional['InputMessageContent'] = None


InputMessageContent = Union['InputTextMessageContent',
                            'InputLocationMessageContent',
                            'InputVenueMessageContent',
                            'InputContactMessageContent']


@attr.s(auto_attribs=True)
class InputTextMessageContent(BaseTelegram):
    message_text: str
    parse_mode: Optional[ParseMode] = None
    entities: Optional[Iterable[MessageEntity]] = None
    disable_web_page_preview: Optional[bool] = None


@attr.s(auto_attribs=True)
class InputLocationMessageContent(BaseTelegram):
    latitude: float
    longitude: float
    horizontal_accuracy: Optional[float] = None
    live_period: Optional[int] = None
    heading: Optional[int] = None
    proximity_alert_radius: Optional[int] = None


@attr.s(auto_attribs=True)
class InputVenueMessageContent(BaseTelegram):
    latitude: float
    longitude: float
    title: str
    address: str
    foursquare_id: Optional[str] = None
    foursquare_type: Optional[str] = None
    google_place_id: Optional[str] = None
    google_place_type: Optional[str] = None


@attr.s(auto_attribs=True)
class InputContactMessageContent(BaseTelegram):
    phone_number: str
    first_name: str
    last_name: Optional[str] = None
    vcard: Optional[str] = None


@attr.s(auto_attribs=True)
class InputInvoiceMessageContent:
    title: str
    description: str
    payload: str
    provider_token: str
    currency: str
    prices: Tuple['LabeledPrice', ...]
    max_tip_amount: Optional[int] = None
    suggested_tip_amounts: Optional[Tuple[int, ...]] = None
    provider_data: Optional[str] = None
    photo_url: Optional[str] = None
    photo_size: Optional[int] = None
    photo_width: Optional[int] = None
    photo_height: Optional[int] = None
    need_name: Optional[bool] = None
    need_phone_number: Optional[bool] = None
    need_email: Optional[bool] = None
    need_shipping_address: Optional[bool] = None
    send_phone_number_to_provider: Optional[bool] = None
    send_email_to_provider: Optional[bool] = None
    is_flexible: Optional[bool] = None


@attr.s(auto_attribs=True)
class ChosenInlineResult(BaseTelegram):
    result_id: str
    from_: User
    query: str
    location: Optional[Location] = None
    inline_message_id: Optional[str] = None


@attr.s(auto_attribs=True)
class LabeledPrice(BaseTelegram):
    label: str
    amount: int


@attr.s(auto_attribs=True)
class Invoice(BaseTelegram):
    title: str
    description: str
    start_parameter: str
    currency: str
    total_amount: int


@attr.s(auto_attribs=True)
class ShippingAddress(BaseTelegram):
    country_code: str
    state: str
    city: str
    street_line1: str
    street_line2: str
    post_code: str


@attr.s(auto_attribs=True)
class OrderInfo(BaseTelegram):
    name: Optional[str] = None
    phone_number: Optional[str] = None
    email: Optional[str] = None
    shipping_address: Optional[ShippingAddress] = None


@attr.s(auto_attribs=True)
class ShippingOption(BaseTelegram):
    id: str
    title: str
    prices: Tuple[LabeledPrice, ...]


@attr.s(auto_attribs=True)
class SuccessfulPayment(BaseTelegram):
    currency: str
    total_amount: int
    invoice_payload: str
    telegram_payment_charge_id: str
    provider_payment_charge_id: str
    shipping_option_id: Optional[str] = None
    order_info: Optional[OrderInfo] = None


@attr.s(auto_attribs=True)
class ShippingQuery(BaseTelegram):
    id: str
    from_: User
    invoice_payload: str
    shipping_address: ShippingAddress


@attr.s(auto_attribs=True)
class PreCheckoutQuery(BaseTelegram):
    id: str
    from_: User
    currency: str
    total_amount: int
    invoice_payload: str
    shipping_option_id: Optional[str] = None
    order_info: Optional[OrderInfo] = None


@attr.s(auto_attribs=True)
class PassportData(BaseTelegram):
    data: Tuple['EncryptedPassportElement', ...]
    credentials: 'EncryptedCredentials'


@attr.s(auto_attribs=True)
class PassportFile(BaseTelegram):
    file_id: str
    file_unique_id: str
    file_date: int
    file_size: Optional[int] = None


@attr.s(auto_attribs=True)
class EncryptedPassportElement(BaseTelegram):
    type: str
    data: Optional[str] = None
    phone_number: Optional[str] = None
    email: Optional[str] = None
    files: Optional[Tuple[PassportFile, ...]] = None
    front_side: Optional[PassportFile] = None
    reverse_side: Optional[PassportFile] = None
    selfie: Optional[PassportFile] = None
    translation: Optional[Tuple[PassportFile, ...]] = None
    hash: Optional[str] = None


@attr.s(auto_attribs=True)
class EncryptedCredentials(BaseTelegram):
    data: str
    hash: str
    secret: str


PassportElementError = Union['PassportElementErrorDataField',
                             'PassportElementErrorFrontSide',
                             'PassportElementErrorReverseSide',
                             'PassportElementErrorSelfie',
                             'PassportElementErrorFile',
                             'PassportElementErrorFiles',
                             'PassportElementErrorTranslationFile',
                             'PassportElementErrorTranslationFiles',
                             'PassportElementErrorUnspecified']


@attr.s(auto_attribs=True)
class PassportElementErrorDataField(BaseTelegram):
    source: str
    type: str
    field_name: str
    data_hash: str
    message: str


@attr.s(auto_attribs=True)
class PassportElementErrorFrontSide(BaseTelegram):
    source: str
    type: str
    file_hash: str
    message: str


@attr.s(auto_attribs=True)
class PassportElementErrorReverseSide(BaseTelegram):
    source: str
    type: str
    file_hash: str
    message: str


@attr.s(auto_attribs=True)
class PassportElementErrorSelfie(BaseTelegram):
    source: str
    type: str
    file_hash: str
    message: str


@attr.s(auto_attribs=True)
class PassportElementErrorFile(BaseTelegram):
    source: str
    type: str
    file_hash: str
    message: str


@attr.s(auto_attribs=True)
class PassportElementErrorFiles(BaseTelegram):
    source: str
    type: str
    file_hashes: List[str]
    message: str


@attr.s(auto_attribs=True)
class PassportElementErrorTranslationFile(BaseTelegram):
    source: str
    type: str
    file_hash: str
    message: str


@attr.s(auto_attribs=True)
class PassportElementErrorTranslationFiles(BaseTelegram):
    source: str
    type: str
    file_hashes: List[str]
    message: str


@attr.s(auto_attribs=True)
class PassportElementErrorUnspecified(BaseTelegram):
    source: str
    type: str
    element_hash: str
    message: str


@attr.s(auto_attribs=True)
class Game(BaseTelegram):
    title: str
    description: str
    photo: Tuple[PhotoSize, ...]
    text: Optional[str] = None
    text_entities: Optional[Tuple[MessageEntity, ...]] = None
    animation: Optional['Animation'] = None


@attr.s(auto_attribs=True)
class CallbackGame(BaseTelegram):
    pass


@attr.s(auto_attribs=True)
class GameHighScore(BaseTelegram):
    position: int
    user: User
    score: int



from abc import abstractmethod
from typing import (Any, AsyncIterator, Dict, List, Protocol, Tuple, Union,
                    runtime_checkable)

__all__ = ('Json', 'StorageProtocol')

Json = Union[str, int, float, bool, Dict[str, Any], List[Any], None]


@runtime_checkable
class StorageProtocol(Protocol):

    @abstractmethod
    async def connect(self) -> None: ...

    @abstractmethod
    async def close(self) -> None: ...

    @abstractmethod
    async def set(self, key: str, value: Json = None) -> None: ...

    @abstractmethod
    async def get(self, key: str) -> Json: ...

    @abstractmethod
    async def delete(self, key: str) -> None: ...

    @abstractmethod
    def iterate(self, prefix: str = '') -> AsyncIterator[Tuple[str, Json]]: ...

    @abstractmethod
    async def clear(self) -> None: ...



import re
from typing import Callable, Final, Iterable, List, Optional, Union

from frozenlist import FrozenList

from .bot import Bot, FilterProtocol, Handler, HandlerCallable
from .bot_update import BotUpdate
from .constants import ContentType, UpdateType
from .filters import (CallbackQueryDataFilter, CommandsFilter,
                      ContentTypeFilter, MessageTextFilter, StateFilter,
                      UpdateTypeFilter)

__all__ = ('HandlerTable',)

HandlerDecorator = Callable[[HandlerCallable], HandlerCallable]


class HandlerTable:

    def __init__(self) -> None:
        self._handlers: Final[FrozenList[Handler]] = FrozenList()

    def freeze(self) -> None:
        self._handlers.freeze()

    @property
    def frozen(self) -> bool:
        return self._handlers.frozen

    async def get_handler(self, bot: Bot,
                          update: BotUpdate) -> Optional[HandlerCallable]:
        for handler in self._handlers:
            if await handler.check(bot, update):
                return handler.callable
        return None

    def message_handler(
        self, handler: HandlerCallable,
        state: Optional[str] = None,
        commands: Optional[Iterable[str]] = None,
        content_types: Optional[Iterable[ContentType]] = None,
        text_match: Union[str, 're.Pattern[str]', None] = None,
        filters: Optional[Iterable[FilterProtocol]] = None
    ) -> None:
        update_type_filter = UpdateTypeFilter(UpdateType.MESSAGE)
        handler_filters: List[FilterProtocol] = [update_type_filter]
        if state is not None:
            handler_filters.append(StateFilter(state))
        if commands is not None:
            handler_filters.append(CommandsFilter(tuple(commands)))
        if content_types is not None:
            handler_filters.append(ContentTypeFilter(tuple(content_types)))
        if isinstance(text_match, re.Pattern):
            handler_filters.append(MessageTextFilter(text_match))
        elif isinstance(text_match, str):
            handler_filters.append(MessageTextFilter(re.compile(text_match)))
        if filters is not None:
            handler_filters.extend(filters)
        self._handlers.append(Handler(handler, tuple(handler_filters)))

    def message(
        self, state: Optional[str] = None,
        commands: Optional[Iterable[str]] = None,
        content_types: Optional[Iterable[ContentType]] = None,
        text_match: Union[str, 're.Pattern[str]', None] = None,
        filters: Optional[Iterable[FilterProtocol]] = None
    ) -> HandlerDecorator:
        def decorator(handler: HandlerCallable) -> HandlerCallable:
            self.message_handler(handler=handler, state=state,
                                 commands=commands,
                                 content_types=content_types,
                                 text_match=text_match, filters=filters)
            return handler
        return decorator

    def edited_message_handler(
        self, handler: HandlerCallable,
        state: Optional[str] = None,
        filters: Optional[Iterable[FilterProtocol]] = None
    ) -> None:
        update_type_filter = UpdateTypeFilter(UpdateType.EDITED_MESSAGE)
        handler_filters: List[FilterProtocol] = [update_type_filter]
        if state is not None:
            handler_filters.append(StateFilter(state))
        if filters is not None:
            handler_filters.extend(filters)
        self._handlers.append(Handler(handler, tuple(handler_filters)))

    def edited_message(
        self, state: Optional[str] = None,
        filters: Optional[Iterable[FilterProtocol]] = None
    ) -> HandlerDecorator:
        def decorator(handler: HandlerCallable) -> HandlerCallable:
            self.edited_message_handler(handler=handler, state=state,
                                        filters=filters)
            return handler
        return decorator

    def channel_post_handler(
        self, handler: HandlerCallable,
        state: Optional[str] = None,
        filters: Optional[Iterable[FilterProtocol]] = None
    ) -> None:
        update_type_filter = UpdateTypeFilter(UpdateType.CHANNEL_POST)
        handler_filters: List[FilterProtocol] = [update_type_filter]
        if state is not None:
            handler_filters.append(StateFilter(state))
        if filters is not None:
            handler_filters.extend(filters)
        self._handlers.append(Handler(handler, tuple(handler_filters)))

    def channel_post(
        self, state: Optional[str] = None,
        filters: Optional[Iterable[FilterProtocol]] = None
    ) -> HandlerDecorator:
        def decorator(handler: HandlerCallable) -> HandlerCallable:
            self.channel_post_handler(handler=handler, state=state,
                                      filters=filters)
            return handler
        return decorator

    def edited_channel_post_handler(
        self, handler: HandlerCallable,
        state: Optional[str] = None,
        filters: Optional[Iterable[FilterProtocol]] = None
    ) -> None:
        update_type_filter = UpdateTypeFilter(UpdateType.EDITED_CHANNEL_POST)
        handler_filters: List[FilterProtocol] = [update_type_filter]
        if state is not None:
            handler_filters.append(StateFilter(state))
        if filters is not None:
            handler_filters.extend(filters)
        self._handlers.append(Handler(handler, tuple(handler_filters)))

    def edited_channel_post(
        self, state: Optional[str] = None,
        filters: Optional[Iterable[FilterProtocol]] = None
    ) -> HandlerDecorator:
        def decorator(handler: HandlerCallable) -> HandlerCallable:
            self.edited_channel_post_handler(handler=handler, state=state,
                                             filters=filters)
            return handler
        return decorator

    def inline_query_handler(
        self, handler: HandlerCallable,
        state: Optional[str] = None,
        filters: Optional[Iterable[FilterProtocol]] = None
    ) -> None:
        update_type_filter = UpdateTypeFilter(UpdateType.INLINE_QUERY)
        handler_filters: List[FilterProtocol] = [update_type_filter]
        if state is not None:
            handler_filters.append(StateFilter(state))
        if filters is not None:
            handler_filters.extend(filters)
        self._handlers.append(Handler(handler, tuple(handler_filters)))

    def inline_query(
        self,
        state: Optional[str] = None,
        filters: Optional[Iterable[FilterProtocol]] = None
    ) -> HandlerDecorator:
        def decorator(handler: HandlerCallable) -> HandlerCallable:
            self.inline_query_handler(handler=handler, state=state,
                                      filters=filters)
            return handler
        return decorator

    def chosen_inline_result_handler(
        self, handler: HandlerCallable,
        state: Optional[str] = None,
        filters: Optional[Iterable[FilterProtocol]] = None
    ) -> None:
        update_type_filter = UpdateTypeFilter(UpdateType.CHOSEN_INLINE_RESULT)
        handler_filters: List[FilterProtocol] = [update_type_filter]
        if state is not None:
            handler_filters.append(StateFilter(state))
        if filters is not None:
            handler_filters.extend(filters)
        self._handlers.append(Handler(handler, tuple(handler_filters)))

    def chosen_inline_result(
        self, state: Optional[str] = None,
        filters: Optional[Iterable[FilterProtocol]] = None
    ) -> HandlerDecorator:
        def decorator(handler: HandlerCallable) -> HandlerCallable:
            self.chosen_inline_result_handler(handler=handler, state=state,
                                              filters=filters)
            return handler
        return decorator

    def callback_query_handler(
        self, handler: HandlerCallable,
        state: Optional[str] = None,
        data_match: Union[str, 're.Pattern[str]', None] = None,
        filters: Optional[Iterable[FilterProtocol]] = None
    ) -> None:
        update_type_filter = UpdateTypeFilter(UpdateType.CALLBACK_QUERY)
        handler_filters: List[FilterProtocol] = [update_type_filter]
        if state is not None:
            handler_filters.append(StateFilter(state))
        if isinstance(data_match, re.Pattern):
            handler_filters.append(CallbackQueryDataFilter(data_match))
        elif isinstance(data_match, str):
            handler_filters.append(CallbackQueryDataFilter(
                re.compile(data_match)))
        if filters is not None:
            handler_filters.extend(filters)
        self._handlers.append(Handler(handler, tuple(handler_filters)))

    def callback_query(
        self, state: Optional[str] = None,
        data_match: Union[str, 're.Pattern[str]', None] = None,
        filters: Optional[Iterable[FilterProtocol]] = None
    ) -> HandlerDecorator:
        def decorator(handler: HandlerCallable) -> HandlerCallable:
            self.callback_query_handler(handler=handler, state=state,
                                        data_match=data_match, filters=filters)
            return handler
        return decorator

    def shipping_query_handler(
        self, handler: HandlerCallable,
        state: Optional[str] = None,
        filters: Optional[Iterable[FilterProtocol]] = None
    ) -> None:
        update_type_filter = UpdateTypeFilter(UpdateType.SHIPPING_QUERY)
        handler_filters: List[FilterProtocol] = [update_type_filter]
        if state is not None:
            handler_filters.append(StateFilter(state))
        if filters is not None:
            handler_filters.extend(filters)
        self._handlers.append(Handler(handler, tuple(handler_filters)))

    def shipping_query(
        self, state: Optional[str] = None,
        filters: Optional[Iterable[FilterProtocol]] = None
    ) -> HandlerDecorator:
        def decorator(handler: HandlerCallable) -> HandlerCallable:
            self.shipping_query_handler(handler=handler, state=state,
                                        filters=filters)
            return handler
        return decorator

    def pre_checkout_query_handler(
        self, handler: HandlerCallable,
        state: Optional[str] = None,
        filters: Optional[Iterable[FilterProtocol]] = None
    ) -> None:
        update_type_filter = UpdateTypeFilter(UpdateType.PRE_CHECKOUT_QUERY)
        handler_filters: List[FilterProtocol] = [update_type_filter]
        if state is not None:
            handler_filters.append(StateFilter(state))
        if filters is not None:
            handler_filters.extend(filters)
        self._handlers.append(Handler(handler, tuple(handler_filters)))

    def pre_checkout_query(
        self, state: Optional[str] = None,
        filters: Optional[Iterable[FilterProtocol]] = None
    ) -> HandlerDecorator:
        def decorator(handler: HandlerCallable) -> HandlerCallable:
            self.pre_checkout_query_handler(handler=handler, state=state,
                                            filters=filters)
            return handler
        return decorator

    def poll_handler(
        self, handler: HandlerCallable,
        state: Optional[str] = None,
        filters: Optional[Iterable[FilterProtocol]] = None
    ) -> None:
        update_type_filter = UpdateTypeFilter(UpdateType.POLL)
        handler_filters: List[FilterProtocol] = [update_type_filter]
        if state is not None:
            handler_filters.append(StateFilter(state))
        if filters is not None:
            handler_filters.extend(filters)
        self._handlers.append(Handler(handler, tuple(handler_filters)))

    def poll(
        self, state: Optional[str] = None,
        filters: Optional[Iterable[FilterProtocol]] = None
    ) -> HandlerDecorator:
        def decorator(handler: HandlerCallable) -> HandlerCallable:
            self.poll_handler(handler=handler, state=state, filters=filters)
            return handler
        return decorator

    def poll_answer_handler(
        self, handler: HandlerCallable,
        state: Optional[str] = None,
        filters: Optional[Iterable[FilterProtocol]] = None
    ) -> None:
        update_type_filter = UpdateTypeFilter(UpdateType.POLL_ANSWER)
        handler_filters: List[FilterProtocol] = [update_type_filter]
        if state is not None:
            handler_filters.append(StateFilter(state))
        if filters is not None:
            handler_filters.extend(filters)
        self._handlers.append(Handler(handler, tuple(handler_filters)))

    def poll_answer(
        self, state: Optional[str] = None,
        filters: Optional[Iterable[FilterProtocol]] = None
    ) -> HandlerDecorator:
        def decorator(handler: HandlerCallable) -> HandlerCallable:
            self.poll_answer_handler(handler=handler, state=state,
                                     filters=filters)
            return handler
        return decorator

    def my_chat_member_handler(
        self, handler: HandlerCallable,
        state: Optional[str] = None,
        filters: Optional[Iterable[FilterProtocol]] = None
    ) -> None:
        update_type_filter = UpdateTypeFilter(UpdateType.MY_CHAT_MEMBER)
        handler_filters: List[FilterProtocol] = [update_type_filter]
        if state is not None:
            handler_filters.append(StateFilter(state))
        if filters is not None:
            handler_filters.extend(filters)
        self._handlers.append(Handler(handler, tuple(handler_filters)))

    def my_chat_member(
        self, state: Optional[str] = None,
        filters: Optional[Iterable[FilterProtocol]] = None
    ) -> HandlerDecorator:
        def decorator(handler: HandlerCallable) -> HandlerCallable:
            self.my_chat_member_handler(handler=handler, state=state,
                                        filters=filters)
            return handler
        return decorator

    def chat_member_handler(
        self, handler: HandlerCallable,
        state: Optional[str] = None,
        filters: Optional[Iterable[FilterProtocol]] = None
    ) -> None:
        update_type_filter = UpdateTypeFilter(UpdateType.CHAT_MEMBER)
        handler_filters: List[FilterProtocol] = [update_type_filter]
        if state is not None:
            handler_filters.append(StateFilter(state))
        if filters is not None:
            handler_filters.extend(filters)
        self._handlers.append(Handler(handler, tuple(handler_filters)))

    def chat_member(
        self, state: Optional[str] = None,
        filters: Optional[Iterable[FilterProtocol]] = None
    ) -> HandlerDecorator:
        def decorator(handler: HandlerCallable) -> HandlerCallable:
            self.chat_member_handler(handler=handler, state=state,
                                     filters=filters)
            return handler
        return decorator



__version__ = '0.10.4'

from .api_types import (BaseTelegram, CallbackQuery, Chat, ChosenInlineResult,
                        Contact, File, InlineKeyboardButton,
                        InlineKeyboardMarkup, InlineQuery, KeyboardButton,
                        LocalFile, Message, PreCheckoutQuery,
                        ReplyKeyboardMarkup, ReplyKeyboardRemove,
                        ShippingQuery, StreamFile, User)
from .bot import Bot, FilterProtocol, PollBot
from .bot_update import BotUpdate
from .constants import (ChatAction, ChatMemberStatus, ChatType, ContentType,
                        ParseMode, PollType, UpdateType)
from .exceptions import (BadGateway, BotBlocked, BotKicked, MigrateToChat,
                         RestartingTelegram, RetryAfter, TelegramError)
from .filters import (ANDFilter, CallbackQueryDataFilter, CommandsFilter,
                      ContentTypeFilter, GroupChatFilter, MessageTextFilter,
                      ORFilter, PrivateChatFilter, StateFilter,
                      UpdateTypeFilter)
from .handler_table import HandlerTable
from .runner import ContextFunction, Runner
from .storage import StorageProtocol

__all__ = (
    '__version__',
    'BaseTelegram',
    'CallbackQuery',
    'Chat',
    'ChosenInlineResult',
    'Contact',
    'File',
    'InlineKeyboardMarkup',
    'InlineKeyboardButton',
    'InlineQuery',
    'KeyboardButton',
    'Message',
    'PreCheckoutQuery',
    'ReplyKeyboardMarkup',
    'ReplyKeyboardRemove',
    'ShippingQuery',
    'User',

    'FilterProtocol',
    'Bot',
    'PollBot',
    'LocalFile',
    'StreamFile',

    'BotUpdate',

    'ChatType',
    'ChatAction',
    'ContentType',
    'ParseMode',
    'PollType',
    'UpdateType',
    'ChatMemberStatus',

    'BadGateway',
    'BotBlocked',
    'BotKicked',
    'MigrateToChat',
    'RestartingTelegram',
    'RetryAfter',
    'TelegramError',

    'CommandsFilter',
    'ContentTypeFilter',
    'GroupChatFilter',
    'PrivateChatFilter',
    'MessageTextFilter',
    'CallbackQueryDataFilter',
    'StateFilter',
    'UpdateTypeFilter',
    'ORFilter',
    'ANDFilter',

    'HandlerTable',

    'StorageProtocol',

    'ContextFunction',
    'Runner'
)



import asyncio
import json
import logging
from ipaddress import IPv4Address, IPv4Network
from secrets import compare_digest, token_urlsafe
from typing import Any, Final, Optional, Tuple, Union

from aiohttp import BaseConnector
from aiohttp.web import (Application, HTTPInternalServerError, HTTPNotFound,
                         Request, Response, StreamResponse)
from yarl import URL

from .api_types import InputFile, Update
from .bot import Bot, HandlerTableProtocol
from .storage import StorageProtocol

NETWORKS: Final[Tuple[IPv4Network, ...]] = (IPv4Network('149.154.160.0/20'),
                                            IPv4Network('91.108.4.0/22'))

bot_logger: Final[logging.Logger] = logging.getLogger('aiotgbot.bot')


class ListenBot(Bot):

    def __init__(
        self,
        url: Union[str, URL],
        token: str,
        handler_table: 'HandlerTableProtocol',
        storage: StorageProtocol,
        certificate: Optional[InputFile] = None,
        ip_address: Optional[str] = None,
        connector: Optional[BaseConnector] = None,
        check_address: bool = False,
        address_header: Optional[str] = None,
        **application_args: Any
    ) -> None:
        super().__init__(token, handler_table, storage, connector)
        self._url: URL = URL(url) if isinstance(url, str) else url
        self._certificate = certificate
        self._ip_address = ip_address
        self._webhook_token: Optional[str] = None
        self._check_address: Final[bool] = check_address
        self._address_header: Final[Optional[str]] = address_header
        self._application = Application(**application_args)
        self._application.router.add_post('/{token}', self._handler)

    @property
    def application(self) -> Application:
        return self._application

    def _address_is_allowed(self, request: Request) -> bool:
        if self._address_header is not None:
            address = IPv4Address(request.headers[self._address_header])
        else:
            address = IPv4Address(request.remote)
        return any(address in network for network in NETWORKS)

    async def _handler(self, request: Request) -> StreamResponse:
        if not self._started:
            raise HTTPInternalServerError()
        assert self._scheduler is not None
        assert self._webhook_token is not None
        if self._check_address and not self._address_is_allowed(request):
            raise HTTPNotFound()
        if not compare_digest(self._webhook_token,
                              request.match_info['token']):
            raise HTTPNotFound()
        update_data = json.loads(await request.read())
        update = Update.from_dict(update_data)
        await self._scheduler.spawn(self._handle_update(update))
        return Response()

    async def start(self) -> None:
        if self._started:
            raise RuntimeError('Polling already started')
        await self._start()
        assert self._me is not None
        loop = asyncio.get_running_loop()
        self._webhook_token = await loop.run_in_executor(None, token_urlsafe)
        assert isinstance(self._webhook_token, str)
        url = str(self._url / self._webhook_token)
        await self.set_webhook(url, self._certificate, self._ip_address)
        bot_logger.info('Bot %s (%s) start listen', self._me.first_name,
                        self._me.username)

    async def stop(self) -> None:
        if not self._started:
            raise RuntimeError('Polling not started')
        if self._stopped:
            raise RuntimeError('Polling already stopped')
        assert self._me is not None
        self._stopped = True
        await self.delete_webhook()
        await self._cleanup()
        bot_logger.info('Bot %s (%s) stop listen', self._me.first_name,
                        self._me.username)



import json
from enum import Enum
from pathlib import Path
from typing import (Any, AsyncIterator, Dict, Final, Optional, Tuple, Union,
                    cast)

import aiosqlite

from .helpers import json_dumps
from .storage import Json, StorageProtocol

__all__ = ('IsolationLevel', 'SQLiteStorage')


class IsolationLevel(str, Enum):
    DEFERRED = 'DEFERRED'
    IMMEDIATE = 'IMMEDIATE'
    EXCLUSIVE = 'EXCLUSIVE'


class SQLiteStorage(StorageProtocol):

    def __init__(self, database: Union[str, Path],
                 isolation_level: Optional[IsolationLevel] = None,
                 **kwargs: Any) -> None:
        self._database: Final[Union[str, Path]] = database
        self._isolation_level: Final[Optional[str]] = (
            isolation_level.value if isolation_level is not None else None)
        self._kwargs: Final[Dict[str, Any]] = kwargs
        self._connection: Optional[aiosqlite.Connection] = None

    async def connect(self) -> None:
        if self._connection is not None:
            raise RuntimeError('Already connected')
        self._connection = await aiosqlite.connect(
            self._database, isolation_level=self._isolation_level,
            **self._kwargs)
        async with self._connection.cursor() as cursor:
            await cursor.execute(
                'CREATE TABLE IF NOT EXISTS kv '
                '(key TEXT NOT NULL PRIMARY KEY, value TEXT NOT NULL)')

    @property
    def connection(self) -> aiosqlite.Connection:
        if self._connection is None:
            raise RuntimeError('Not connected')
        return self._connection

    async def close(self) -> None:
        if self._connection is None:
            raise RuntimeError('Not connected')
        await self._connection.close()
        self._connection = None

    async def set(self, key: str, value: Json = None) -> None:
        async with self.connection.cursor() as cursor:
            await cursor.execute(
                'INSERT OR REPLACE INTO kv (key, value) VALUES (?, ?)',
                (key, json_dumps(value)))

    async def get(self, key: str) -> Json:
        async with self.connection.cursor() as cursor:
            await cursor.execute('SELECT value FROM kv WHERE key = ?', (key,))
            row = await cursor.fetchone()
            if row is not None:
                return cast(Json, json.loads(row[0]))
            else:
                return None

    async def delete(self, key: str) -> None:
        async with self.connection.cursor() as cursor:
            await cursor.execute('DELETE FROM kv WHERE key = ?', (key,))

    async def iterate(
        self, prefix: str = ''
    ) -> AsyncIterator[Tuple[str, Json]]:
        async with self.connection.execute(
            'SELECT key, value FROM kv WHERE key LIKE ? ORDER BY key',
            (f'{prefix}%',)
        ) as cursor:
            async for row in cursor:
                yield row[0], json.loads(row[1])

    async def clear(self) -> None:
        async with self.connection.cursor() as cursor:
            await cursor.execute('DELETE FROM kv')
            await cursor.execute('VACUUM')



from typing import Any, Dict, Final, Iterator, MutableMapping, Optional

from .api_types import (CallbackQuery, ChatMemberUpdated, ChosenInlineResult,
                        InlineQuery, Message, Poll, PollAnswer,
                        PreCheckoutQuery, ShippingQuery, Update)

__all__ = ('Context', 'BotUpdate')


class Context(MutableMapping[str, Any]):

    def __init__(self, data: Dict[str, Any]) -> None:
        self._data: Final[Dict[str, Any]] = data

    def __getitem__(self, key: str) -> Any:
        return self._data[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self._data[key] = value

    def __delitem__(self, key: str) -> None:
        del self._data[key]

    def __len__(self) -> int:
        return len(self._data)

    def __iter__(self) -> Iterator[str]:
        return iter(self._data)

    def clear(self) -> None:
        self._data.clear()

    def to_dict(self) -> Dict[str, Any]:
        return self._data


class BotUpdate(MutableMapping[str, Any]):

    def __init__(self, state: Optional[str], context: Context,
                 update: Update) -> None:
        self._state: Optional[str] = state
        self._context: Final[Context] = context
        self._update: Final[Update] = update
        self._data: Final[Dict[str, Any]] = {}

    def __getitem__(self, key: str) -> Any:
        return self._data[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self._data[key] = value

    def __delitem__(self, key: str) -> None:
        del self._data[key]

    def __len__(self) -> int:
        return len(self._data)

    def __iter__(self) -> Iterator[str]:
        return iter(self._data)

    @property
    def state(self) -> Optional[str]:
        return self._state

    @state.setter
    def state(self, value: str) -> None:
        self._state = value

    @property
    def context(self) -> Context:
        return self._context

    @property
    def update_id(self) -> int:
        return self._update.update_id

    @property
    def message(self) -> Optional[Message]:
        return self._update.message

    @property
    def edited_message(self) -> Optional[Message]:
        return self._update.edited_message

    @property
    def channel_post(self) -> Optional[Message]:
        return self._update.channel_post

    @property
    def edited_channel_post(self) -> Optional[Message]:
        return self._update.edited_channel_post

    @property
    def inline_query(self) -> Optional[InlineQuery]:
        return self._update.inline_query

    @property
    def chosen_inline_result(self) -> Optional[ChosenInlineResult]:
        return self._update.chosen_inline_result

    @property
    def callback_query(self) -> Optional[CallbackQuery]:
        return self._update.callback_query

    @property
    def shipping_query(self) -> Optional[ShippingQuery]:
        return self._update.shipping_query

    @property
    def pre_checkout_query(self) -> Optional[PreCheckoutQuery]:
        return self._update.pre_checkout_query

    @property
    def poll(self) -> Optional[Poll]:
        return self._update.poll

    @property
    def poll_answer(self) -> Optional[PollAnswer]:
        return self._update.poll_answer

    @property
    def my_chat_member(self) -> Optional[ChatMemberUpdated]:
        return self._update.my_chat_member

    @property
    def chat_member(self) -> Optional[ChatMemberUpdated]:
        return self._update.chat_member



import asyncio
import json
import logging
from abc import abstractmethod
from functools import partial
from http import HTTPStatus
from typing import (Any, Awaitable, Callable, Dict, Final, Iterator,
                    MutableMapping, Optional, Protocol, Tuple, Union,
                    runtime_checkable)

import aiojobs
import attr
import backoff
from aiofreqlimit import FreqLimit
from aiohttp import (BaseConnector, ClientError, ClientSession, FormData,
                     TCPConnector)
from aiojobs_protocols import SchedulerProtocol

from .api_methods import ApiMethods, ParamType
from .api_types import APIResponse, LocalFile, StreamFile, Update, User
from .bot_update import BotUpdate, Context
from .constants import ChatType, RequestMethod
from .exceptions import (BadGateway, BotBlocked, BotKicked, MigrateToChat,
                         RestartingTelegram, RetryAfter, TelegramError)
from .helpers import KeyLock, get_software, json_dumps
from .storage import StorageProtocol

__all__ = ('Bot', 'PollBot', 'HandlerCallable', 'Handler',
           'HandlerTableProtocol', 'FilterProtocol')

SOFTWARE: Final[str] = get_software()
TG_API_URL: Final[str] = 'https://api.telegram.org/bot{token}/{method}'
TG_FILE_URL: Final[str] = 'https://api.telegram.org/file/bot{token}/{path}'
TG_GET_UPDATES_TIMEOUT: Final[int] = 60
STATE_PREFIX: Final[str] = 'state'
CONTEXT_PREFIX: Final[str] = 'context'
MESSAGE_INTERVAL: Final[float] = 1 / 30
CHAT_INTERVAL: Final[float] = 1
GROUP_INTERVAL: Final[float] = 3

bot_logger: Final[logging.Logger] = logging.getLogger('aiotgbot.bot')
response_logger: Final[logging.Logger] = logging.getLogger('aiotgbot.response')

EventHandler = Callable[['Bot'], Awaitable[None]]


class Bot(MutableMapping[str, Any], ApiMethods):

    def __init__(
        self,
        token: str,
        handler_table: 'HandlerTableProtocol',
        storage: StorageProtocol,
        connector: Optional[BaseConnector] = None
    ) -> None:
        if not handler_table.frozen:
            raise RuntimeError('Can\'t use unfrozen handler table')
        if not isinstance(handler_table, HandlerTableProtocol):
            raise RuntimeError('Handler table is not HandlerTableProtocol '
                               'instance')
        if not isinstance(storage, StorageProtocol):
            raise RuntimeError('Storage parameter is not StorageProtocol '
                               'instance')
        self._token: Final[str] = token
        self._handler_table: Final['HandlerTableProtocol'] = handler_table
        self._storage: Final[StorageProtocol] = storage
        if connector is not None:
            _connector: BaseConnector = connector
        else:
            _connector = TCPConnector(keepalive_timeout=60)
        self._client: Final[ClientSession] = ClientSession(
            connector=_connector,
            json_serialize=json_dumps,
            headers={'User-Agent': SOFTWARE}
        )
        self._context_lock: Final[KeyLock] = KeyLock()
        self._message_limit: Final[FreqLimit] = FreqLimit(MESSAGE_INTERVAL)
        self._chat_limit: Final[FreqLimit] = FreqLimit(CHAT_INTERVAL)
        self._group_limit: Final[FreqLimit] = FreqLimit(GROUP_INTERVAL)
        self._scheduler: Optional[SchedulerProtocol] = None
        self._started: bool = False
        self._stopped: bool = False
        self._updates_offset: int = 0
        self._me: Optional[User] = None
        self._data: Final[Dict[str, Any]] = {}

    def __getitem__(self, key: str) -> Any:
        return self._data[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self._data[key] = value

    def __delitem__(self, key: str) -> None:
        del self._data[key]

    def __len__(self) -> int:
        return len(self._data)

    def __iter__(self) -> Iterator[str]:
        return iter(self._data)

    @property
    def id(self) -> int:
        return int(self._token.split(':')[0])

    @property
    def storage(self) -> StorageProtocol:
        return self._storage

    @property
    def client(self) -> ClientSession:
        if self._client is None:
            raise RuntimeError('Access to client during bot is not running.')
        else:
            return self._client

    def file_url(self, path: str) -> str:
        return TG_FILE_URL.format(token=self._token, path=path)

    @staticmethod
    def _scheduler_exception_handler(_: SchedulerProtocol,
                                     context: Dict[str, Any]) -> None:
        bot_logger.exception('Update handle error',
                             exc_info=context['exception'])

    @staticmethod
    def _telegram_exception(api_response: APIResponse) -> TelegramError:
        assert api_response.error_code is not None
        assert api_response.description is not None
        error_code = api_response.error_code
        description = api_response.description
        if (
            api_response.parameters is not None and
            api_response.parameters.retry_after is not None
        ):
            retry_after = api_response.parameters.retry_after
            assert retry_after is not None
            return RetryAfter(error_code, description, retry_after)
        elif (api_response.parameters is not None and
              api_response.parameters.migrate_to_chat_id is not None):
            return MigrateToChat(error_code, description,
                                 api_response.parameters.migrate_to_chat_id)
        elif (error_code >= HTTPStatus.INTERNAL_SERVER_ERROR and
              RestartingTelegram.match(description)):
            return RestartingTelegram(error_code, description)
        elif BadGateway.match(description):
            return BadGateway(error_code, description)
        elif BotBlocked.match(description):
            return BotBlocked(error_code, description)
        elif BotKicked.match(description):
            return BotKicked(error_code, description)
        else:
            return TelegramError(error_code, description)

    @backoff.on_exception(backoff.expo, ClientError)
    async def _request(
            self,
            http_method: RequestMethod,
            api_method: str,
            **params: ParamType
    ) -> APIResponse:
        data = {
            name: str(value) if isinstance(value, (int,  float)) else value
            for name, value in params.items() if value is not None
        }
        bot_logger.debug('Request %s %s %r', http_method, api_method, data)
        if http_method == RequestMethod.GET:
            if len(data) > 0:
                assert all(isinstance(value, str) for value in data.values())
                request = partial(self.client.get, params=data)
            else:
                request = partial(self.client.get)
        else:
            form_data = FormData()
            for name, value in data.items():
                if isinstance(value, (StreamFile, LocalFile)):
                    form_data.add_field(name, value.content,
                                        content_type=value.content_type,
                                        filename=value.name)
                else:
                    form_data.add_field(name, value)
            request = partial(self.client.post, data=form_data)

        url = TG_API_URL.format(token=self._token, method=api_method)
        async with request(url) as response:
            response_dict = json.loads(await response.read())
        response_logger.debug(response_dict)
        api_response = APIResponse.from_dict(response_dict)

        if api_response.ok:
            return api_response
        else:
            raise Bot._telegram_exception(api_response)

    async def _safe_request(
            self,
            http_method: RequestMethod,
            api_method: str,
            chat_id: Union[int, str],
            **params: ParamType
    ) -> APIResponse:
        retry_allowed = all(not isinstance(param, StreamFile)
                            for param in params.values())
        request = partial(self._request, http_method, api_method,
                          chat_id=chat_id, **params)

        chat = await self.get_chat(chat_id)
        while True:
            try:
                message_limit = self._message_limit.acquire()
                if chat.type in (ChatType.GROUP, ChatType.SUPERGROUP):
                    group_limit = self._group_limit.acquire(chat.id)
                    async with message_limit, group_limit:
                        return await request()
                else:
                    chat_limit = self._chat_limit.acquire(chat.id)
                    async with message_limit, chat_limit:
                        return await request()
            except RetryAfter as retry_after:
                if retry_allowed:
                    await asyncio.sleep(retry_after.retry_after)
                else:
                    bot_logger.error('RetryAfter error during '
                                     'retry not allowed')
                    raise

    @staticmethod
    def _update_state(update: Update) -> str:
        user_id: Optional[int] = None
        chat_id: Optional[int] = None

        if update.message is not None:
            assert update.message.from_ is not None
            user_id = update.message.from_.id
            assert update.message.chat is not None
            chat_id = update.message.chat.id
        elif update.edited_message is not None:
            assert update.edited_message.from_ is not None
            user_id = update.edited_message.from_.id
            assert update.edited_message.chat is not None
            chat_id = update.edited_message.chat.id
        elif update.channel_post is not None:
            assert update.channel_post.chat is not None
            chat_id = update.channel_post.chat.id
        elif update.edited_channel_post is not None:
            assert update.edited_channel_post.chat is not None
            chat_id = update.edited_channel_post.chat.id
        elif update.inline_query is not None:
            user_id = update.inline_query.from_.id
        elif update.chosen_inline_result is not None:
            user_id = update.chosen_inline_result.from_.id
        elif (update.callback_query is not None and
              update.callback_query.message is not None):
            user_id = update.callback_query.from_.id
            assert update.callback_query.message.chat is not None
            chat_id = update.callback_query.message.chat.id
        elif update.callback_query is not None:
            user_id = update.callback_query.from_.id
        elif update.shipping_query is not None:
            user_id = update.shipping_query.from_.id
        elif update.pre_checkout_query is not None:
            user_id = update.pre_checkout_query.from_.id
        elif update.poll is not None:
            pass
        elif update.poll_answer is not None:
            user_id = update.poll_answer.user.id
        elif update.my_chat_member is not None:
            assert update.my_chat_member.from_ is not None
            user_id = update.my_chat_member.from_.id
            assert update.my_chat_member.chat is not None
            chat_id = update.my_chat_member.chat.id
        elif update.chat_member is not None:
            assert update.chat_member.from_ is not None
            user_id = update.chat_member.from_.id
            assert update.chat_member.chat is not None
            chat_id = update.chat_member.chat.id

        return '{}|{}'.format(str(user_id) if user_id is not None else '',
                              str(chat_id) if chat_id is not None else '')

    async def _handle_update(self, update: Update) -> None:
        assert self._handler_table.frozen
        assert self._context_lock is not None, 'Context lock not initialized'
        bot_logger.debug('Dispatch update "%s"', update.update_id)
        update_state = self._update_state(update)
        state_key = f'{STATE_PREFIX}|{update_state}'
        context_key = f'{CONTEXT_PREFIX}|{update_state}'
        async with self._context_lock.acquire(state_key):
            state = await self._storage.get(state_key)
            assert isinstance(state, str) or state is None
            context_dict = await self._storage.get(context_key)
            assert isinstance(context_dict, dict) or context_dict is None
            context = Context(context_dict if context_dict is not None else {})
            bot_update = BotUpdate(state, context, update)
            handler = await self._handler_table.get_handler(self, bot_update)
            if handler is not None:
                bot_logger.debug('Dispatched update "%s" to "%s"',
                                 update.update_id,
                                 handler.__name__)
                await handler(self, bot_update)
                await self._storage.set(state_key, bot_update.state)
                await self._storage.set(context_key,
                                        bot_update.context.to_dict())
                bot_logger.debug('Set context for update "%s"',
                                 update.update_id)
            else:
                bot_logger.debug('Not found handler for update "%s". Skip.',
                                 update.update_id)

    async def _start(self) -> None:
        self._started = True

        self._me = await self.get_me()
        self._scheduler = await aiojobs.create_scheduler(
            exception_handler=self._scheduler_exception_handler)

    async def _cleanup(self) -> None:
        assert self._client is not None
        assert self._scheduler is not None
        await self._scheduler.close()
        await self._client.close()
        await self._message_limit.clear()
        await self._chat_limit.clear()
        await self._group_limit.clear()

    @abstractmethod
    async def start(self) -> None: ...

    @abstractmethod
    async def stop(self) -> None: ...


class PollBot(Bot):

    def __init__(
        self,
        token: str,
        handler_table: 'HandlerTableProtocol',
        storage: StorageProtocol,
        connector: Optional[BaseConnector] = None
    ) -> None:
        super().__init__(token, handler_table, storage, connector)
        self._poll_task: Optional[asyncio.Task[None]] = None

    async def start(self) -> None:
        if self._started:
            raise RuntimeError('Polling already started')
        await self._start()
        assert self._me is not None
        self._poll_task = asyncio.create_task(self._poll_wrapper())
        bot_logger.info('Bot %s (%s) start polling', self._me.first_name,
                        self._me.username)

    async def stop(self) -> None:
        if not self._started:
            raise RuntimeError('Polling not started')
        if self._stopped:
            raise RuntimeError('Polling already stopped')
        assert self._poll_task is not None
        bot_logger.debug('Stop polling')
        self._stopped = True
        if not self._poll_task.done():
            self._poll_task.cancel()

    async def _poll_wrapper(self) -> None:
        assert self._me is not None
        try:
            await self._poll()
        except asyncio.CancelledError:
            pass
        except Exception as exception:
            bot_logger.exception('Error while polling updates',
                                 exc_info=exception)
        await self._cleanup()
        bot_logger.info('Bot %s (%s) stop polling', self._me.first_name,
                        self._me.username)

    @backoff.on_exception(backoff.expo, TelegramError)
    async def _poll(self) -> None:
        assert self._scheduler is not None, 'Scheduler not initialized'
        bot_logger.debug('Get updates from: %s', self._updates_offset)
        while not self._stopped:
            updates = await self.get_updates(offset=self._updates_offset,
                                             timeout=TG_GET_UPDATES_TIMEOUT)
            for update in updates:
                await self._scheduler.spawn(self._handle_update(update))
            if len(updates) > 0:
                self._updates_offset = updates[-1].update_id + 1


HandlerCallable = Callable[[Bot, BotUpdate], Awaitable[None]]
FiltersType = Tuple['FilterProtocol', ...]


@attr.s(frozen=True, auto_attribs=True)
class Handler:
    callable: HandlerCallable
    filters: FiltersType

    async def check(self, bot: Bot, update: BotUpdate) -> bool:
        return all([await _filter.check(bot, update)
                    for _filter in self.filters])


@runtime_checkable
class HandlerTableProtocol(Protocol):

    def freeze(self) -> None: ...

    @property
    def frozen(self) -> bool: ...  # noqa

    @abstractmethod
    async def get_handler(
        self, bot: Bot, update: BotUpdate
    ) -> Optional[HandlerCallable]: ...


@runtime_checkable
class FilterProtocol(Protocol):

    @abstractmethod
    async def check(self, bot: Bot, update: BotUpdate) -> bool: ...



import logging
from abc import ABC, abstractmethod
from enum import Enum
from itertools import count
from typing import Any, Dict, Final, Iterable, Optional, Tuple, Union

import attr

from .api_types import (APIResponse, BaseTelegram, BotCommand, BotCommandScope,
                        Chat, ChatInviteLink, ChatMember, ChatPermissions,
                        File, GameHighScore, InlineKeyboardMarkup,
                        InlineQueryResult, InputFile, InputMedia,
                        InputMediaAudio, InputMediaDocument, InputMediaPhoto,
                        InputMediaVideo, LabeledPrice, MaskPosition, Message,
                        MessageEntity, MessageId, PassportElementError, Poll,
                        ReplyMarkup, ShippingOption, StickerSet, Update, User,
                        UserProfilePhotos, WebhookInfo)
from .constants import (ChatAction, DiceEmoji, ParseMode, PollType,
                        RequestMethod, UpdateType)
from .helpers import json_dumps

__all__ = ('ParamType', 'ApiMethods')

api_logger: Final[logging.Logger] = logging.getLogger('aiotgbot.api')


_ItemType = Union[BaseTelegram, str, int, Enum]


def _unstructure_item(
    item: Union[BaseTelegram, str, int, Enum]
) -> Union[Dict[str, Any], str, int]:
    if isinstance(item, BaseTelegram):
        return item.to_dict()
    elif isinstance(item, Enum) and isinstance(item.value, str):
        return item.value
    elif isinstance(item, Enum):
        raise TypeError(f'Unsupported enum type: {item!r}')
    elif isinstance(item, (str, int)):
        return item
    else:
        raise TypeError(f'Unsupported item type: {item!r}')


def _json_dumps(
    value: Union[BaseTelegram, Iterable[_ItemType], None]
) -> Optional[str]:
    if value is None:
        return None
    elif isinstance(value, Iterable):
        return json_dumps(tuple(_unstructure_item(item) for item in value))
    elif isinstance(value, BaseTelegram):
        return json_dumps(value.to_dict())
    else:
        raise TypeError(f'Unsupported value type: {value!r}')


def _enum_to_str(item: Optional[Enum]) -> Optional[str]:
    assert item is None or isinstance(item.value, str)
    return item.value if item is not None else None


ParamType = Union[int, float, str, InputFile, None]


class ApiMethods(ABC):

    @abstractmethod
    async def _request(self, http_method: RequestMethod, api_method: str,
                       **params: ParamType) -> APIResponse: ...

    @abstractmethod
    async def _safe_request(
            self, http_method: RequestMethod, api_method: str,
            chat_id: Union[int, str], **params: ParamType
    ) -> APIResponse: ...

    async def get_updates(
        self, offset: Optional[int] = None,
        limit: Optional[int] = None,
        timeout: Optional[int] = None,
        allowed_updates: Optional[Iterable[UpdateType]] = None
    ) -> Tuple[Update, ...]:
        api_logger.debug(f'Get updates offset: {offset}, limit: {limit}, '
                         f'timeout: {timeout}, '
                         f'allowed_updates: {allowed_updates}')
        response = await self._request(
            RequestMethod.GET, 'getUpdates', offset=offset, limit=limit,
            timeout=timeout, allowed_updates=_json_dumps(allowed_updates))

        return tuple(Update.from_dict(item) for item in response.result)

    async def set_webhook(
        self, url: Optional[str] = None,
        certificate: Optional[InputFile] = None,
        ip_address: Optional[str] = None,
        max_connections: Optional[int] = None,
        allowed_updates: Optional[Iterable[UpdateType]] = None,
        drop_pending_updates: Optional[bool] = None
    ) -> bool:
        api_logger.debug('Set webhook')
        response = await self._request(
            RequestMethod.POST, 'setWebhook', url=url, certificate=certificate,
            ip_address=ip_address, max_connections=max_connections,
            allowed_updates=_json_dumps(allowed_updates),
            drop_pending_updates=drop_pending_updates)
        assert isinstance(response.result, bool)
        return response.result

    async def delete_webhook(
        self, drop_pending_updates: Optional[bool] = None
    ) -> bool:
        api_logger.debug('Delete webhook')
        response = await self._request(
            RequestMethod.POST, 'deleteWebhook',
            drop_pending_updates=drop_pending_updates)
        assert isinstance(response.result, bool)
        return response.result

    async def get_webhook_info(self) -> WebhookInfo:
        api_logger.debug('Get webhook info')
        response = await self._request(RequestMethod.GET, 'getWebhookInfo')

        return WebhookInfo.from_dict(response.result)

    async def get_me(self) -> User:
        api_logger.debug('Get me')
        response = await self._request(RequestMethod.GET, 'getMe')
        return User.from_dict(response.result)

    async def log_out(self) -> bool:
        api_logger.debug('Log out')
        response = await self._request(RequestMethod.POST, 'logOut')
        assert isinstance(response.result, bool)
        return response.result

    async def close(self) -> bool:
        api_logger.debug('Close')
        response = await self._request(RequestMethod.POST, 'close')
        assert isinstance(response.result, bool)
        return response.result

    async def send_message(
        self, chat_id: Union[int, str], text: str,
        parse_mode: Optional[ParseMode] = None,
        entities: Optional[Iterable[MessageEntity]] = None,
        disable_web_page_preview: Optional[bool] = None,
        disable_notification: Optional[bool] = None,
        reply_to_message_id: Optional[int] = None,
        allow_sending_without_reply: Optional[bool] = None,
        reply_markup: Optional[ReplyMarkup] = None
    ) -> Message:
        api_logger.debug('Send message %r to chat "%s"', text, chat_id)
        response = await self._safe_request(
            RequestMethod.POST, 'sendMessage', chat_id, text=text,
            parse_mode=_enum_to_str(parse_mode),
            entities=_json_dumps(entities),
            disable_web_page_preview=disable_web_page_preview,
            disable_notification=disable_notification,
            reply_to_message_id=reply_to_message_id,
            allow_sending_without_reply=allow_sending_without_reply,
            reply_markup=_json_dumps(reply_markup))

        return Message.from_dict(response.result)

    async def forward_message(
        self, chat_id: Union[int, str], from_chat_id: Union[int, str],
        message_id: int, disable_notification: Optional[bool] = None
    ) -> Message:
        api_logger.debug('Forward message %s to "%s" from "%s"', message_id,
                         chat_id, from_chat_id)
        response = await self._safe_request(
            RequestMethod.POST, 'forwardMessage', chat_id,
            from_chat_id=from_chat_id, message_id=message_id,
            disable_notification=disable_notification)

        return Message.from_dict(response.result)

    async def copy_message(
        self, chat_id: Union[int, str], from_chat_id: Union[int, str],
        message_id: int, caption: Optional[str] = None,
        parse_mode: Optional[ParseMode] = None,
        caption_entities: Optional[Iterable[MessageEntity]] = None,
        disable_notification: Optional[bool] = None,
        reply_to_message_id: Optional[int] = None,
        allow_sending_without_reply: Optional[bool] = None,
        reply_markup: Optional[ReplyMarkup] = None
    ) -> MessageId:
        api_logger.debug('Copy message %s to "%s" from "%s"', message_id,
                         chat_id, from_chat_id)
        response = await self._safe_request(
            RequestMethod.POST, 'copyMessage', chat_id,
            from_chat_id=from_chat_id, message_id=message_id,
            parse_mode=parse_mode,
            caption_entities=_json_dumps(caption_entities),
            caption=caption, disable_notification=disable_notification,
            reply_to_message_id=reply_to_message_id,
            allow_sending_without_reply=allow_sending_without_reply,
            reply_markup=_json_dumps(reply_markup))

        return MessageId.from_dict(response.result)

    async def send_photo(
        self, chat_id: Union[int, str],
        photo: Union[InputFile, str],
        caption: Optional[str] = None,
        parse_mode: Optional[ParseMode] = None,
        caption_entities: Optional[Iterable[MessageEntity]] = None,
        disable_notification: Optional[bool] = None,
        reply_to_message_id: Optional[int] = None,
        allow_sending_without_reply: Optional[bool] = None,
        reply_markup: Optional[ReplyMarkup] = None
    ) -> Message:
        api_logger.debug('Send photo to "%s"', chat_id)
        response = await self._safe_request(
            RequestMethod.POST, 'sendPhoto', chat_id, photo=photo,
            caption=caption, parse_mode=_enum_to_str(parse_mode),
            caption_entities=_json_dumps(caption_entities),
            disable_notification=disable_notification,
            reply_to_message_id=reply_to_message_id,
            allow_sending_without_reply=allow_sending_without_reply,
            reply_markup=_json_dumps(reply_markup))

        return Message.from_dict(response.result)

    async def send_audio(
        self, chat_id: Union[int, str],
        audio: Union[InputFile, str],
        caption: Optional[str] = None,
        parse_mode: Optional[ParseMode] = None,
        caption_entities: Optional[Iterable[MessageEntity]] = None,
        disable_notification: Optional[bool] = None,
        duration: Optional[int] = None,
        performer: Optional[str] = None,
        title: Optional[str] = None,
        thumb: Optional[Union[InputFile, str]] = None,
        reply_to_message_id: Optional[int] = None,
        allow_sending_without_reply: Optional[bool] = None,
        reply_markup: Optional[ReplyMarkup] = None
    ) -> Message:
        api_logger.debug('Send audio to "%s"', chat_id)
        response = await self._safe_request(
            RequestMethod.POST, 'sendAudio', chat_id, audio=audio,
            caption=caption, parse_mode=_enum_to_str(parse_mode),
            caption_entities=_json_dumps(caption_entities),
            disable_notification=disable_notification,
            duration=duration, performer=performer,
            title=title, thumb=thumb,
            reply_to_message_id=reply_to_message_id,
            allow_sending_without_reply=allow_sending_without_reply,
            reply_markup=_json_dumps(reply_markup))

        return Message.from_dict(response.result)

    async def send_document(
        self, chat_id: Union[int, str],
        document: Union[InputFile, str],
        thumb: Optional[Union[InputFile, str]],
        caption: Optional[str] = None,
        parse_mode: Optional[ParseMode] = None,
        caption_entities: Optional[Iterable[MessageEntity]] = None,
        disable_content_type_detection: Optional[bool] = None,
        disable_notification: Optional[bool] = None,
        duration: Optional[int] = None,
        performer: Optional[str] = None,
        title: Optional[str] = None,
        reply_to_message_id: Optional[int] = None,
        allow_sending_without_reply: Optional[bool] = None,
        reply_markup: Optional[ReplyMarkup] = None
    ) -> Message:
        api_logger.debug('Send document to "%s"', chat_id)
        response = await self._safe_request(
            RequestMethod.POST, 'sendDocument', chat_id,
            document=document, thumb=thumb, caption=caption,
            parse_mode=_enum_to_str(parse_mode),
            caption_entities=_json_dumps(caption_entities),
            disable_content_type_detection=disable_content_type_detection,
            disable_notification=disable_notification,
            duration=duration, performer=performer, title=title,
            reply_to_message_id=reply_to_message_id,
            allow_sending_without_reply=allow_sending_without_reply,
            reply_markup=_json_dumps(reply_markup))

        return Message.from_dict(response.result)

    async def send_video(
        self, chat_id: Union[int, str],
        video: Union[InputFile, str],
        duration: Optional[int] = None,
        width: Optional[int] = None,
        height: Optional[int] = None,
        thumb: Optional[Union[InputFile, str]] = None,
        caption: Optional[str] = None,
        parse_mode: Optional[ParseMode] = None,
        caption_entities: Optional[Iterable[MessageEntity]] = None,
        supports_streaming: Optional[bool] = None,
        disable_notification: Optional[bool] = None,
        reply_to_message_id: Optional[int] = None,
        allow_sending_without_reply: Optional[bool] = None,
        reply_markup: Optional[ReplyMarkup] = None
    ) -> Message:
        api_logger.debug('Send video to "%s"', chat_id)
        response = await self._safe_request(
            RequestMethod.POST, 'sendVideo', chat_id,
            video=video, duration=duration, width=width,
            height=height, thumb=thumb, caption=caption,
            parse_mode=_enum_to_str(parse_mode),
            caption_entities=_json_dumps(caption_entities),
            supports_streaming=supports_streaming,
            disable_notification=disable_notification,
            reply_to_message_id=reply_to_message_id,
            allow_sending_without_reply=allow_sending_without_reply,
            reply_markup=_json_dumps(reply_markup))

        return Message.from_dict(response.result)

    async def send_animation(
        self, chat_id: Union[int, str],
        animation: Union[InputFile, str],
        duration: Optional[int] = None,
        width: Optional[int] = None,
        height: Optional[int] = None,
        thumb: Optional[Union[InputFile, str]] = None,
        caption: Optional[str] = None,
        parse_mode: Optional[ParseMode] = None,
        caption_entities: Optional[Iterable[MessageEntity]] = None,
        disable_notification: Optional[bool] = None,
        reply_to_message_id: Optional[int] = None,
        allow_sending_without_reply: Optional[bool] = None,
        reply_markup: Optional[ReplyMarkup] = None
    ) -> Message:
        api_logger.debug('Send animation to "%s"', chat_id)
        response = await self._safe_request(
            RequestMethod.POST, 'sendAnimation', chat_id,
            animation=animation, duration=duration, width=width,
            height=height, thumb=thumb, caption=caption,
            parse_mode=_enum_to_str(parse_mode),
            caption_entities=_json_dumps(caption_entities),
            disable_notification=disable_notification,
            reply_to_message_id=reply_to_message_id,
            allow_sending_without_reply=allow_sending_without_reply,
            reply_markup=_json_dumps(reply_markup))

        return Message.from_dict(response.result)

    async def send_voice(
        self, chat_id: Union[int, str],
        voice: Union[InputFile, str],
        caption: Optional[str] = None,
        parse_mode: Optional[ParseMode] = None,
        caption_entities: Optional[Iterable[MessageEntity]] = None,
        duration: Optional[int] = None,
        disable_notification: Optional[bool] = None,
        reply_to_message_id: Optional[int] = None,
        allow_sending_without_reply: Optional[bool] = None,
        reply_markup: Optional[ReplyMarkup] = None
    ) -> Message:
        api_logger.debug('Send voice to "%s"', chat_id)
        response = await self._safe_request(
            RequestMethod.POST, 'sendVoice', chat_id,
            voice=voice, caption=caption,
            parse_mode=_enum_to_str(parse_mode),
            caption_entities=_json_dumps(caption_entities),
            duration=duration,
            disable_notification=disable_notification,
            reply_to_message_id=reply_to_message_id,
            allow_sending_without_reply=allow_sending_without_reply,
            reply_markup=_json_dumps(reply_markup))

        return Message.from_dict(response.result)

    async def send_video_note(
        self, chat_id: Union[int, str],
        video_note: Union[InputFile, str],
        duration: Optional[int] = None,
        length: Optional[int] = None,
        thumb: Optional[Union[InputFile, str]] = None,
        disable_notification: Optional[bool] = None,
        reply_to_message_id: Optional[int] = None,
        allow_sending_without_reply: Optional[bool] = None,
        reply_markup: Optional[ReplyMarkup] = None
    ) -> Message:
        api_logger.debug('Send video not to "%s"', chat_id)
        response = await self._safe_request(
            RequestMethod.POST, 'sendVideoNote', chat_id,
            video_note=video_note, duration=duration,
            length=length, thumb=thumb,
            disable_notification=disable_notification,
            reply_to_message_id=reply_to_message_id,
            allow_sending_without_reply=allow_sending_without_reply,
            reply_markup=_json_dumps(reply_markup))

        return Message.from_dict(response.result)

    async def send_media_group(
        self, chat_id: Union[int, str],
        media: Iterable[Union[InputMediaAudio, InputMediaDocument,
                              InputMediaPhoto, InputMediaVideo]],
        disable_notification: Optional[bool] = None,
        reply_to_message_id: Optional[int] = None,
        allow_sending_without_reply: Optional[bool] = None
    ) -> Tuple[Message, ...]:
        api_logger.debug('Send media group to "%s"', chat_id)
        attached_media = []
        attachments = {}
        counter = count()
        for item in media:
            if isinstance(item.media, str):
                attached_media.append(item)
            else:
                attachment_name = f'attachment{next(counter)}'
                attachments[attachment_name] = item.media
                attached_media.append(attr.evolve(
                    item, media=f'attach://{attachment_name}'))
        response = await self._safe_request(
            RequestMethod.POST, 'sendMediaGroup', chat_id,
            media=_json_dumps(attached_media),
            disable_notification=disable_notification,
            reply_to_message_id=reply_to_message_id,
            allow_sending_without_reply=allow_sending_without_reply,
            **attachments
        )

        return tuple(Message.from_dict(item) for item in response.result)

    async def send_location(
        self, chat_id: Union[int, str],
        latitude: float,
        longitude: float,
        horizontal_accuracy: Optional[float] = None,
        live_period: Optional[int] = None,
        heading: Optional[int] = None,
        proximity_alert_radius: Optional[int] = None,
        length: Optional[int] = None,
        disable_notification: Optional[bool] = None,
        reply_to_message_id: Optional[int] = None,
        allow_sending_without_reply: Optional[bool] = None,
        reply_markup: Optional[ReplyMarkup] = None
    ) -> Message:
        api_logger.debug('Send location to "%s"', chat_id)
        response = await self._safe_request(
            RequestMethod.POST, 'sendLocation', chat_id,
            latitude=latitude, longitude=longitude,
            horizontal_accuracy=horizontal_accuracy,
            live_period=live_period, heading=heading,
            proximity_alert_radius=proximity_alert_radius, length=length,
            disable_notification=disable_notification,
            reply_to_message_id=reply_to_message_id,
            allow_sending_without_reply=allow_sending_without_reply,
            reply_markup=_json_dumps(reply_markup))

        return Message.from_dict(response.result)

    async def edit_message_live_location(
        self,
        latitude: float,
        longitude: float,
        horizontal_accuracy: Optional[float] = None,
        heading: Optional[int] = None,
        proximity_alert_radius: Optional[int] = None,
        chat_id: Optional[Union[int, str]] = None,
        message_id: Optional[int] = None,
        inline_message_id: Optional[str] = None,
        reply_markup: Optional[ReplyMarkup] = None
    ) -> Union[Message, bool]:
        api_logger.debug('Edit live location %s in "%s"', message_id, chat_id)
        response = await self._request(
            RequestMethod.POST, 'editMessageLiveLocation',
            chat_id=chat_id, message_id=message_id,
            inline_message_id=inline_message_id,
            latitude=latitude, longitude=longitude,
            horizontal_accuracy=horizontal_accuracy, heading=heading,
            proximity_alert_radius=proximity_alert_radius,
            reply_markup=_json_dumps(reply_markup))

        if isinstance(response.result, bool):
            return response.result
        else:
            return Message.from_dict(response.result)

    async def stop_message_live_location(
        self, chat_id: Optional[Union[int, str]] = None,
        message_id: Optional[int] = None,
        inline_message_id: Optional[str] = None,
        reply_markup: Optional[ReplyMarkup] = None
    ) -> Union[Message, bool]:
        api_logger.debug('Stop live location %s in "%s"', message_id, chat_id)
        response = await self._request(
            RequestMethod.POST, 'stopMessageLiveLocation',
            chat_id=chat_id, message_id=message_id,
            inline_message_id=inline_message_id,
            reply_markup=_json_dumps(reply_markup))

        if isinstance(response.result, bool):
            return response.result
        else:
            return Message.from_dict(response.result)

    async def send_venue(
        self, chat_id: Union[int, str],
        latitude: float,
        longitude: float,
        title: str,
        address: str,
        foursquare_id: Optional[str] = None,
        foursquare_type: Optional[str] = None,
        disable_notification: Optional[bool] = None,
        reply_to_message_id: Optional[int] = None,
        allow_sending_without_reply: Optional[bool] = None,
        reply_markup: Optional[ReplyMarkup] = None
    ) -> Message:
        api_logger.debug('Send venue to "%s"', chat_id)
        response = await self._safe_request(
            RequestMethod.POST, 'sendVenue', chat_id,
            latitude=latitude, longitude=longitude, title=title,
            address=address, foursquare_id=foursquare_id,
            foursquare_type=foursquare_type,
            disable_notification=disable_notification,
            reply_to_message_id=reply_to_message_id,
            allow_sending_without_reply=allow_sending_without_reply,
            reply_markup=_json_dumps(reply_markup))

        return Message.from_dict(response.result)

    async def send_contact(
        self, chat_id: Union[int, str],
        phone_number: str,
        first_name: str,
        last_name: Optional[str] = None,
        vcard: Optional[str] = None,
        disable_notification: Optional[bool] = None,
        reply_to_message_id: Optional[int] = None,
        allow_sending_without_reply: Optional[bool] = None,
        reply_markup: Optional[ReplyMarkup] = None
    ) -> Message:
        api_logger.debug('Send contact to "%s"', chat_id)
        response = await self._safe_request(
            RequestMethod.POST, 'sendContact', chat_id,
            phone_number=phone_number, first_name=first_name,
            last_name=last_name, vcard=vcard,
            disable_notification=disable_notification,
            reply_to_message_id=reply_to_message_id,
            allow_sending_without_reply=allow_sending_without_reply,
            reply_markup=_json_dumps(reply_markup))

        return Message.from_dict(response.result)

    async def send_poll(
        self, chat_id: Union[int, str], question: str,
        options: Iterable[str], is_anonymous: Optional[bool] = None,
        type_: Optional[PollType] = None,
        allows_multiple_answers: Optional[bool] = None,
        correct_option_id: Optional[int] = None,
        is_closed: Optional[bool] = None,
        explanation: Optional[str] = None,
        explanation_parse_mode: Optional[ParseMode] = None,
        explanation_entities: Optional[Iterable[MessageEntity]] = None,
        open_period: Optional[int] = None,
        close_date: Optional[int] = None,
        disable_notification: Optional[bool] = None,
        reply_to_message_id: Optional[int] = None,
        allow_sending_without_reply: Optional[bool] = None,
        reply_markup: Optional[ReplyMarkup] = None
    ) -> Message:
        api_logger.debug('Send poll to chat "%s"', chat_id)
        response = await self._safe_request(
            RequestMethod.POST, 'sendPoll', chat_id,
            question=question, options=_json_dumps(options),
            is_anonymous=is_anonymous, type=type_,
            allows_multiple_answers=allows_multiple_answers,
            correct_option_id=correct_option_id,
            is_closed=is_closed, explanation=explanation,
            explanation_parse_mode=explanation_parse_mode,
            explanation_entities=_json_dumps(explanation_entities),
            open_period=open_period, close_date=close_date,
            disable_notification=disable_notification,
            reply_to_message_id=reply_to_message_id,
            allow_sending_without_reply=allow_sending_without_reply,
            reply_markup=_json_dumps(reply_markup))

        return Message.from_dict(response.result)

    async def send_dice(self, chat_id: Union[int, str],
                        emoji: Optional[DiceEmoji] = None,
                        disable_notification: Optional[bool] = None,
                        reply_to_message_id: Optional[int] = None,
                        allow_sending_without_reply: Optional[bool] = None,
                        reply_markup: Optional[ReplyMarkup] = None) -> Message:
        api_logger.debug('Send dice "%s" to chat "%s"', emoji, chat_id)
        response = await self._safe_request(
            RequestMethod.POST, 'sendDice', chat_id,
            emoji=_enum_to_str(emoji),
            disable_notification=disable_notification,
            reply_to_message_id=reply_to_message_id,
            allow_sending_without_reply=allow_sending_without_reply,
            reply_markup=_json_dumps(reply_markup))

        return Message.from_dict(response.result)

    async def send_chat_action(self, chat_id: Union[int, str],
                               action: ChatAction) -> bool:
        api_logger.debug('Send action "%s" to chat "%s"', action, chat_id)
        response = await self._safe_request(
            RequestMethod.POST, 'sendChatAction', chat_id, action=action.value)
        assert isinstance(response.result, bool)
        return response.result

    async def get_user_profile_photos(
        self, user_id: int,
        offset: Optional[int] = None,
        limit: Optional[int] = None
    ) -> UserProfilePhotos:
        api_logger.debug('Get user profile photos %s offset %s limit %s',
                         user_id, offset, limit)
        response = await self._request(
            RequestMethod.GET, 'getUserProfilePhotos',
            user_id=user_id, offset=offset, limit=limit)

        return UserProfilePhotos.from_dict(response.result)

    async def get_file(self, file_id: str) -> File:
        api_logger.debug('Get file "%s"', file_id)
        response = await self._request(RequestMethod.GET, 'getFile',
                                       file_id=file_id)
        return File.from_dict(response.result)

    async def ban_chat_member(self, chat_id: Union[int, str], user_id: int,
                              until_date: Optional[int] = None,
                              revoke_messages: Optional[bool] = None) -> bool:
        response = await self._request(
            RequestMethod.POST, 'banChatMember',
            chat_id=chat_id, user_id=user_id, until_date=until_date,
            revoke_messages=revoke_messages)
        assert isinstance(response.result, bool)
        return response.result

    async def unban_chat_member(self, chat_id: Union[int, str], user_id: int,
                                only_if_banned: Optional[bool] = None) -> bool:
        api_logger.debug('Unban member %s in "%s"', user_id, chat_id)
        response = await self._request(RequestMethod.POST, 'unbanChatMember',
                                       chat_id=chat_id, user_id=user_id,
                                       only_if_banned=only_if_banned)
        assert isinstance(response.result, bool)
        return response.result

    async def restrict_chat_member(
        self, chat_id: Union[int, str],
        user_id: int,
        permissions: ChatPermissions,
        until_date: Optional[int] = None,
    ) -> bool:
        api_logger.debug('Restrict member %s in "%s"', user_id, chat_id)
        response = await self._request(
            RequestMethod.POST, 'restrictChatMember', chat_id=chat_id,
            user_id=user_id, permissions=_json_dumps(permissions),
            until_date=until_date)
        assert isinstance(response.result, bool)
        return response.result

    async def promote_chat_member(
        self, chat_id: Union[int, str],
        user_id: int,
        is_anonymous: Optional[int] = None,
        can_manage_chat: Optional[int] = None,
        can_change_info: Optional[int] = None,
        can_post_messages: Optional[bool] = None,
        can_edit_messages: Optional[bool] = None,
        can_delete_messages: Optional[bool] = None,
        can_manage_voice_chats: Optional[bool] = None,
        can_invite_users: Optional[bool] = None,
        can_restrict_members: Optional[bool] = None,
        can_pin_messages: Optional[bool] = None,
        can_promote_members: Optional[bool] = None
    ) -> bool:
        api_logger.debug('Promote member %s in "%s"', user_id, chat_id)
        response = await self._request(
            RequestMethod.POST, 'promoteChatMember',
            chat_id=chat_id, user_id=user_id, is_anonymous=is_anonymous,
            can_manage_chat=can_manage_chat,
            can_change_info=can_change_info,
            can_post_messages=can_post_messages,
            can_edit_messages=can_edit_messages,
            can_delete_messages=can_delete_messages,
            can_manage_voice_chats=can_manage_voice_chats,
            can_invite_users=can_invite_users,
            can_restrict_members=can_restrict_members,
            can_pin_messages=can_pin_messages,
            can_promote_members=can_promote_members)
        assert isinstance(response.result, bool)
        return response.result

    async def set_chat_administrator_custom_title(
        self, chat_id: Union[int, str],
        user_id: int,
        custom_title: str
    ) -> bool:
        api_logger.debug('Set title "%s" for admin %s in "%s"',
                         custom_title, user_id, chat_id)
        response = await self._request(
            RequestMethod.POST, 'setChatAdministratorCustomTitle',
            chat_id=chat_id, user_id=user_id, custom_title=custom_title)
        assert isinstance(response.result, bool)
        return response.result

    async def export_chat_invite_link(self, chat_id: Union[int, str]) -> str:
        api_logger.debug('Export chat "%s" invite link', chat_id)
        response = await self._request(
            RequestMethod.POST, 'exportChatInviteLink', chat_id=chat_id)
        assert isinstance(response.result, str)
        return response.result

    async def create_chat_invite_link(
        self, chat_id: Union[int, str], expire_date: Optional[int] = None,
        member_limit: Optional[int] = None
    ) -> ChatInviteLink:
        api_logger.debug('Create chat "%s" invite link', chat_id)
        response = await self._request(
            RequestMethod.POST, 'createChatInviteLink', chat_id=chat_id,
            expire_date=expire_date, member_limit=member_limit)
        return ChatInviteLink.from_dict(response.result)

    async def edit_chat_invite_link(
        self, chat_id: Union[int, str], invite_link: str,
        expire_date: Optional[int] = None,
        member_limit: Optional[int] = None
    ) -> ChatInviteLink:
        api_logger.debug('Edit chat "%s" invite link', chat_id)
        response = await self._request(
            RequestMethod.POST, 'editChatInviteLink', chat_id=chat_id,
            invite_link=invite_link, expire_date=expire_date,
            member_limit=member_limit)
        return ChatInviteLink.from_dict(response.result)

    async def revoke_chat_invite_link(
        self, chat_id: Union[int, str], invite_link: str
    ) -> ChatInviteLink:
        api_logger.debug('Revoke chat "%s" invite link', chat_id)
        response = await self._request(
            RequestMethod.POST, 'revokeChatInviteLink', chat_id=chat_id,
            invite_link=invite_link)
        return ChatInviteLink.from_dict(response.result)

    async def set_chat_permissions(
        self, chat_id: Union[int, str],
        permissions: ChatPermissions
    ) -> bool:
        api_logger.debug('Set chat "%s" permissions', chat_id)
        response = await self._safe_request(
            RequestMethod.POST, 'setChatPermissions', chat_id,
            permissions=_json_dumps(permissions))
        assert isinstance(response.result, bool)
        return response.result

    async def set_chat_photo(
        self, chat_id: Union[int, str],
        photo: InputFile
    ) -> bool:
        api_logger.debug('Set chat "%s" photo', chat_id)
        response = await self._safe_request(RequestMethod.POST, 'setChatPhoto',
                                            chat_id, photo=photo)
        assert isinstance(response.result, bool)
        return response.result

    async def delete_chat_photo(self, chat_id: Union[int, str]) -> bool:
        api_logger.debug('Delete chat "%s" photo', chat_id)
        response = await self._request(RequestMethod.POST, 'deleteChatPhoto',
                                       chat_id=chat_id)
        assert isinstance(response.result, bool)
        return response.result

    async def set_chat_title(self, chat_id: Union[int, str],
                             title: str) -> bool:
        api_logger.debug('Set title "%s" for chat "%s"', title, chat_id)
        response = await self._request(RequestMethod.POST, 'setChatTitle',
                                       chat_id=chat_id, title=title)
        assert isinstance(response.result, bool)
        return response.result

    async def set_chat_description(self, chat_id: Union[int, str],
                                   description: str) -> bool:
        api_logger.debug('Set chat "%s" description', chat_id)
        response = await self._request(
            RequestMethod.POST, 'setChatDescription', chat_id=chat_id,
            description=description)
        assert isinstance(response.result, bool)
        return response.result

    async def pin_chat_message(
        self, chat_id: Union[int, str], message_id: int,
        disable_notification: Optional[bool] = None
    ) -> bool:
        api_logger.debug('Pin message %s in chat "%s"', message_id, chat_id)
        response = await self._request(
            RequestMethod.POST, 'pinChatMessage',
            chat_id=chat_id, message_id=message_id,
            disable_notification=disable_notification)
        assert isinstance(response.result, bool)
        return response.result

    async def unpin_chat_message(self, chat_id: Union[int, str],
                                 message_id: Optional[int]) -> bool:
        api_logger.debug('Unpin message "%s" in chat "%s"',
                         message_id, chat_id)
        response = await self._request(RequestMethod.POST, 'unpinChatMessage',
                                       chat_id=chat_id, message_id=message_id)
        assert isinstance(response.result, bool)
        return response.result

    async def unpin_all_chat_messages(self, chat_id: Union[int, str]) -> bool:
        api_logger.debug('Unpin all messages in chat "%s"', chat_id)
        response = await self._request(RequestMethod.POST,
                                       'unpinAllChatMessages', chat_id=chat_id)
        assert isinstance(response.result, bool)
        return response.result

    async def leave_chat(self, chat_id: Union[int, str]) -> bool:
        api_logger.debug('Leave chat "%s"', chat_id)
        response = await self._request(RequestMethod.POST, 'leaveChat',
                                       chat_id=chat_id)
        assert isinstance(response.result, bool)
        return response.result

    async def get_chat(self, chat_id: Union[int, str]) -> Chat:
        api_logger.debug('Get chat "%s"', chat_id)
        response = await self._request(RequestMethod.GET, 'getChat',
                                       chat_id=chat_id)
        return Chat.from_dict(response.result)

    async def get_chat_administrators(
        self, chat_id: Union[int, str]
    ) -> Tuple[ChatMember, ...]:
        api_logger.debug('Get chat administrators "%s"', chat_id)
        response = await self._request(
            RequestMethod.GET, 'getChatAdministrators', chat_id=chat_id)

        return tuple(ChatMember.from_dict(item) for item in response.result)

    async def get_chat_member_count(self, chat_id: Union[int, str]) -> int:
        api_logger.debug('Get chat member count "%s"', chat_id)
        response = await self._request(
            RequestMethod.GET, 'getChatMemberCount', chat_id=chat_id)
        assert isinstance(response.result, int)
        return response.result

    async def get_chat_member(self, chat_id: Union[int, str],
                              user_id: int) -> ChatMember:
        api_logger.debug('Get chat "%s" member %s', chat_id, user_id)
        response = await self._request(RequestMethod.GET, 'getChatMember',
                                       chat_id=chat_id, user_id=user_id)

        return ChatMember.from_dict(response.result)

    async def set_chat_sticker_set(self, chat_id: Union[int, str],
                                   sticker_set_name: str) -> bool:
        api_logger.debug('Set chat "%s" sticker set "%s"', chat_id,
                         sticker_set_name)
        response = await self._request(
            RequestMethod.POST, 'setChatStickerSet', chat_id=chat_id,
            sticker_set_name=sticker_set_name)
        assert isinstance(response.result, bool)
        return response.result

    async def delete_chat_sticker_set(self, chat_id: Union[int, str]) -> bool:
        api_logger.debug('Delete chat "%s" sticker set', chat_id)
        response = await self._request(RequestMethod.POST,
                                       'deleteChatStickerSet', chat_id=chat_id)
        assert isinstance(response.result, bool)
        return response.result

    async def answer_callback_query(self, callback_query_id: str,
                                    text: Optional[str] = None,
                                    show_alert: Optional[bool] = None,
                                    url: Optional[str] = None,
                                    cache_time: Optional[int] = None) -> bool:
        api_logger.debug('Answer callback query "%s"', callback_query_id)
        response = await self._request(
            RequestMethod.POST, 'answerCallbackQuery',
            callback_query_id=callback_query_id, text=text,
            show_alert=show_alert, url=url, cache_time=cache_time)
        assert isinstance(response.result, bool)
        return response.result

    async def set_my_commands(self, commands: Iterable[BotCommand],
                              scope: Optional[BotCommandScope] = None,
                              language_code: Optional[str] = None) -> bool:
        api_logger.debug('Set my commands "%s"', commands)
        response = await self._request(
            RequestMethod.POST, 'setMyCommands',
            commands=_json_dumps(commands), scope=_json_dumps(scope),
            language_code=language_code)
        assert isinstance(response.result, bool)
        return response.result

    async def get_my_commands(
        self, scope: Optional[BotCommandScope] = None,
        language_code: Optional[str] = None
    ) -> Tuple[BotCommand, ...]:
        api_logger.debug('Get my commands')
        response = await self._request(
            RequestMethod.GET, 'getMyCommands', scope=_json_dumps(scope),
            language_code=language_code)

        return tuple(BotCommand.from_dict(item) for item in response.result)

    async def delete_my_commands(
        self, scope: Optional[BotCommandScope] = None,
        language_code: Optional[str] = None
    ) -> bool:
        api_logger.debug('Delete my commands')
        response = await self._request(
            RequestMethod.GET, 'deleteMyCommands', scope=_json_dumps(scope),
            language_code=language_code)
        assert isinstance(response.result, bool)
        return response.result

    async def edit_message_text(
        self,
        text: str,
        chat_id: Optional[Union[int, str]] = None,
        message_id: Optional[int] = None,
        inline_message_id: Optional[str] = None,
        parse_mode: Optional[ParseMode] = None,
        entities: Optional[Iterable[MessageEntity]] = None,
        disable_web_page_preview: Optional[bool] = None,
        reply_markup: Optional[InlineKeyboardMarkup] = None
    ) -> Union[Message, bool]:
        if (
            (chat_id is None or message_id is None) and
            inline_message_id is None
        ):
            raise RuntimeError('chat_id or message_id and '
                               'inline_message_id is None')
        if inline_message_id is None:
            api_logger.debug('Edit message %s in "%s" text',
                             message_id, chat_id)
        else:
            api_logger.debug('Edit inline message "%s" text',
                             inline_message_id)
        response = await self._request(
            RequestMethod.POST, 'editMessageText',
            chat_id=chat_id, message_id=message_id,
            inline_message_id=inline_message_id, text=text,
            parse_mode=_enum_to_str(parse_mode),
            entities=_json_dumps(entities),
            disable_web_page_preview=disable_web_page_preview,
            reply_markup=_json_dumps(reply_markup))

        if isinstance(response.result, bool):
            return response.result
        else:
            return Message.from_dict(response.result)

    async def edit_message_caption(
        self,
        chat_id: Optional[Union[int, str]] = None,
        message_id: Optional[int] = None,
        inline_message_id: Optional[str] = None,
        caption: Optional[str] = None,
        parse_mode: Optional[ParseMode] = None,
        caption_entities: Optional[Iterable[MessageEntity]] = None,
        reply_markup: Optional[InlineKeyboardMarkup] = None
    ) -> Union[Message, bool]:
        if (
            (chat_id is None or message_id is None) and
            inline_message_id is None
        ):
            raise RuntimeError('chat_id or message_id and '
                               'inline_message_id is None')
        if inline_message_id is None:
            api_logger.debug('Edit message %s in "%s" caption',
                             message_id, chat_id)
        else:
            api_logger.debug('Edit inline message "%s" caption',
                             inline_message_id)
        response = await self._request(
            RequestMethod.POST, 'editMessageCaption',
            chat_id=chat_id, message_id=message_id,
            inline_message_id=inline_message_id,
            caption=caption, parse_mode=_enum_to_str(parse_mode),
            caption_entities=_json_dumps(caption_entities),
            reply_markup=_json_dumps(reply_markup))

        if isinstance(response.result, bool):
            return response.result
        else:
            return Message.from_dict(response.result)

    async def edit_message_media(
        self, media: InputMedia,
        chat_id: Optional[Union[int, str]] = None,
        message_id: Optional[int] = None,
        inline_message_id: Optional[str] = None,
        reply_markup: Optional[InlineKeyboardMarkup] = None
    ) -> Union[Message, bool]:
        if (
            (chat_id is None or message_id is None) and
            inline_message_id is None
        ):
            raise RuntimeError('chat_id or message_id and '
                               'inline_message_id is None')
        if inline_message_id is None:
            api_logger.debug('Edit message %s in "%s" media',
                             message_id, chat_id)
        else:
            api_logger.debug('Edit inline message "%s" media',
                             inline_message_id)
        attachments = {}
        if not isinstance(media.media, str):
            attachment_name = 'attachment0'
            attachments[attachment_name] = media.media
            media = attr.evolve(media, media=f'attach://{attachment_name}')
        response = await self._request(
            RequestMethod.POST, 'editMessageMedia', chat_id=chat_id,
            message_id=message_id, inline_message_id=inline_message_id,
            media=_json_dumps(media), reply_markup=_json_dumps(reply_markup),
            **attachments)
        if isinstance(response.result, bool):
            return response.result
        else:
            return Message.from_dict(response.result)

    async def edit_message_reply_markup(
        self, chat_id: Optional[Union[int, str]] = None,
        message_id: Optional[int] = None,
        inline_message_id: Optional[str] = None,
        reply_markup: Optional[InlineKeyboardMarkup] = None
    ) -> Union[Message, bool]:
        if (
            (chat_id is None or message_id is None) and
            inline_message_id is None
        ):
            raise RuntimeError('chat_id or message_id and '
                               'inline_message_id is None')
        if inline_message_id is None:
            api_logger.debug('Edit message %s in "%s" reply markup',
                             message_id, chat_id)
        else:
            api_logger.debug('Edit inline message "%s" reply markup',
                             inline_message_id)
        response = await self._request(
            RequestMethod.POST, 'editMessageReplyMarkup',
            chat_id=chat_id, message_id=message_id,
            inline_message_id=inline_message_id,
            reply_markup=_json_dumps(reply_markup))

        if isinstance(response.result, bool):
            return response.result
        else:
            return Message.from_dict(response.result)

    async def stop_poll(
        self, chat_id: Union[int, str], message_id: int,
        reply_markup: Optional[InlineKeyboardMarkup] = None
    ) -> Poll:
        api_logger.debug('Stop poll %s in "%s"', message_id, chat_id)
        response = await self._request(
            RequestMethod.POST, 'stopPoll',
            chat_id=chat_id, message_id=message_id,
            reply_markup=_json_dumps(reply_markup))

        return Poll.from_dict(response.result)

    async def delete_message(self, chat_id: Optional[Union[int, str]] = None,
                             message_id: Optional[int] = None) -> bool:
        api_logger.debug('Delete message %s in "%s"', message_id, chat_id)
        response = await self._request(RequestMethod.POST, 'deleteMessage',
                                       chat_id=chat_id, message_id=message_id)
        assert isinstance(response.result, bool)
        return response.result

    async def send_sticker(
        self, chat_id: Union[int, str],
        sticker: Union[InputFile, str],
        disable_notification: Optional[bool] = None,
        reply_to_message_id: Optional[int] = None,
        allow_sending_without_reply: Optional[bool] = None,
        reply_markup: Optional[ReplyMarkup] = None
    ) -> Message:
        api_logger.debug('Send sticker to "%s"', chat_id)
        response = await self._request(
            RequestMethod.POST, 'sendSticker',
            chat_id=chat_id, sticker=sticker,
            disable_notification=disable_notification,
            reply_to_message_id=reply_to_message_id,
            allow_sending_without_reply=allow_sending_without_reply,
            reply_markup=_json_dumps(reply_markup))

        return Message.from_dict(response.result)

    async def get_sticker_set(self, name: str) -> StickerSet:
        api_logger.debug('Get sticker set "%s"', name)
        response = await self._request(RequestMethod.GET, 'getStickerSet',
                                       name=name)

        return StickerSet.from_dict(response.result)

    async def upload_sticker_file(self, user_id: int,
                                  png_sticker: InputFile) -> File:
        api_logger.debug('Upload sticker file for %s', user_id)
        response = await self._request(
            RequestMethod.POST, 'uploadStickerFile',
            user_id=user_id, png_sticker=png_sticker)

        return File.from_dict(response.result)

    async def create_new_sticker_set(
        self, user_id: int, name: str, title: str, emojis: str,
        png_sticker: Union[InputFile, str, None] = None,
        tgs_sticker: Optional[InputFile] = None,
        contains_masks: Optional[bool] = None,
        mask_position: Optional[MaskPosition] = None
    ) -> bool:
        api_logger.debug('Create new sticker set "%s" for %s', name, user_id)
        response = await self._request(
            RequestMethod.POST, 'createNewStickerSet',
            user_id=user_id, name=name, title=title, emojis=emojis,
            png_sticker=png_sticker, tgs_sticker=tgs_sticker,
            contains_masks=contains_masks,
            mask_position=_json_dumps(mask_position))
        assert isinstance(response.result, bool)
        return response.result

    async def add_sticker_to_set(
        self, user_id: int, name: str, title: str,
        png_sticker: Union[InputFile, str], emojis: str,
        mask_position: Optional[MaskPosition] = None
    ) -> File:
        api_logger.debug('Add sticker to set "%s" for %s', name, user_id)
        response = await self._request(
            RequestMethod.POST, 'addStickerToSet',
            user_id=user_id, name=name, title=title, png_sticker=png_sticker,
            emojis=emojis, mask_position=_json_dumps(mask_position))

        return File.from_dict(response.result)

    async def set_sticker_position_in_set(self, sticker: str,
                                          position: int) -> bool:
        api_logger.debug('Set sticker "%s" position to %s', sticker, position)
        response = await self._request(
            RequestMethod.POST, 'setStickerPositionInSet',
            sticker=sticker, position=position)
        assert isinstance(response.result, bool)
        return response.result

    async def delete_sticker_from_set(self, sticker: str) -> bool:
        api_logger.debug('Delete sticker "%s" from set', sticker)
        response = await self._request(
            RequestMethod.POST, 'deleteStickerFromSet', sticker=sticker)
        assert isinstance(response.result, bool)
        return response.result

    async def set_sticker_set_thumb(
            self, name: str, user_id: int,
            thumb: Union[InputFile, str, None] = None
    ) -> bool:
        api_logger.debug('Set sticker set "%s" owned by "%s" thumb',
                         name, user_id)
        response = await self._request(
            RequestMethod.POST, 'setStickerSetThumb',
            name=name, user_id=user_id, thumb=thumb)
        assert isinstance(response.result, bool)
        return response.result

    async def answer_inline_query(
        self, inline_query_id: str, results: Iterable[InlineQueryResult],
        cache_time: Optional[int] = None, is_personal: Optional[bool] = None,
        next_offset: Optional[str] = None,
        switch_pm_text: Optional[str] = None,
        switch_pm_parameter: Optional[str] = None
    ) -> bool:
        api_logger.debug('Answer inline query "%s"', inline_query_id)
        response = await self._request(
            RequestMethod.POST, 'answerInlineQuery',
            inline_query_id=inline_query_id,
            results=_json_dumps(results),
            cache_time=cache_time, is_personal=is_personal,
            next_offset=next_offset, switch_pm_text=switch_pm_text,
            switch_pm_parameter=switch_pm_parameter)
        assert isinstance(response.result, bool)
        return response.result

    async def send_invoice(
        self, chat_id: int, title: str, description: str, payload: str,
        provider_token: str, currency: str, prices: Iterable[LabeledPrice],
        max_tip_amount: Optional[int] = None,
        suggested_tip_amounts: Optional[Tuple[int, ...]] = None,
        start_parameter: Optional[str] = None,
        provider_data: Optional[str] = None,
        photo_url: Optional[str] = None, photo_size: Optional[int] = None,
        photo_width: Optional[int] = None, photo_height: Optional[int] = None,
        need_name: Optional[bool] = None,
        need_phone_number: Optional[bool] = None,
        need_email: Optional[bool] = None,
        need_shipping_address: Optional[bool] = None,
        send_phone_number_to_provider: Optional[bool] = None,
        send_email_to_provider: Optional[bool] = None,
        is_flexible: Optional[bool] = None,
        disable_notification: Optional[bool] = None,
        reply_to_message_id: Optional[int] = None,
        allow_sending_without_reply: Optional[bool] = None,
        reply_markup: Optional[InlineKeyboardMarkup] = None
    ) -> Message:
        api_logger.debug('Send invoice to %s', chat_id)
        response = await self._safe_request(
            RequestMethod.POST, 'sendInvoice', chat_id,
            title=title, description=description, payload=payload,
            provider_token=provider_token,
            currency=currency,
            prices=_json_dumps(prices),
            max_tip_amount=max_tip_amount,
            suggested_tip_amounts=_json_dumps(suggested_tip_amounts),
            start_parameter=start_parameter,
            provider_data=provider_data, photo_url=photo_url,
            photo_size=photo_size, photo_width=photo_width,
            photo_height=photo_height, need_name=need_name,
            need_phone_number=need_phone_number, need_email=need_email,
            need_shipping_address=need_shipping_address,
            send_phone_number_to_provider=send_phone_number_to_provider,
            send_email_to_provider=send_email_to_provider,
            is_flexible=is_flexible,
            disable_notification=disable_notification,
            reply_to_message_id=reply_to_message_id,
            allow_sending_without_reply=allow_sending_without_reply,
            reply_markup=_json_dumps(reply_markup))

        return Message.from_dict(response.result)

    async def answer_shipping_query(
        self, inline_query_id: str, ok: bool,
        shipping_options: Optional[Iterable[ShippingOption]] = None,
        error_message: Optional[str] = None
    ) -> bool:
        api_logger.debug('Answer shipping query "%s"', inline_query_id)
        response = await self._request(
            RequestMethod.POST, 'answerShippingQuery',
            inline_query_id=inline_query_id, ok=ok,
            shipping_options=_json_dumps(shipping_options),
            error_message=error_message)
        assert isinstance(response.result, bool)
        return response.result

    async def answer_pre_checkout_query(
        self, pre_checkout_query_id: str, ok: bool,
        error_message: Optional[str] = None
    ) -> bool:
        api_logger.debug('Answer pre checkout query "%s"',
                         pre_checkout_query_id)
        response = await self._request(
            RequestMethod.POST, 'answerPreCheckoutQuery',
            pre_checkout_query_id=pre_checkout_query_id, ok=ok,
            error_message=error_message)
        assert isinstance(response.result, bool)
        return response.result

    async def set_passport_data_errors(
        self, user_id: int,
        errors: Iterable[PassportElementError]
    ) -> bool:
        api_logger.debug('Set passport data errors %s', user_id)
        response = await self._request(
            RequestMethod.POST, 'setPassportDataErrors',
            user_id=user_id,
            errors=_json_dumps(errors))
        assert isinstance(response.result, bool)
        return response.result

    async def send_game(
        self, chat_id: int, game_short_name: str,
        disable_notification: Optional[bool] = None,
        reply_to_message_id: Optional[int] = None,
        allow_sending_without_reply: Optional[bool] = None,
        reply_markup: Optional[InlineKeyboardMarkup] = None
    ) -> Message:
        api_logger.debug('Send game "%s" to %s', chat_id, game_short_name)
        response = await self._safe_request(
            RequestMethod.POST, 'sendGame', chat_id,
            game_short_name=game_short_name,
            disable_notification=disable_notification,
            reply_to_message_id=reply_to_message_id,
            allow_sending_without_reply=allow_sending_without_reply,
            reply_markup=_json_dumps(reply_markup))

        return Message.from_dict(response.result)

    async def set_game_score(
        self, user_id: int, score: int,
        force: Optional[bool] = None,
        disable_edit_message: Optional[bool] = None,
        chat_id: Optional[int] = None, message_id: Optional[int] = None,
        inline_message_id: Optional[str] = None
    ) -> Union[Message, bool]:
        api_logger.debug('Set game score %s for %s', score, user_id)
        response = await self._request(
            RequestMethod.POST, 'setGameScore',
            user_id=user_id, score=score, force=force,
            disable_edit_message=disable_edit_message,
            chat_id=chat_id, message_id=message_id,
            inline_message_id=inline_message_id)

        if isinstance(response.result, bool):
            return response.result
        else:
            return Message.from_dict(response.result)

    async def get_game_high_scores(
        self, user_id: int, chat_id: Optional[int] = None,
        message_id: Optional[int] = None,
        inline_message_id: Optional[str] = None
    ) -> Tuple[GameHighScore, ...]:
        api_logger.debug('Get game high scores for %s', user_id)
        response = await self._request(
            RequestMethod.POST, 'getGameHighScores',
            user_id=user_id, chat_id=chat_id, message_id=message_id,
            inline_message_id=inline_message_id)

        return tuple(GameHighScore.from_dict(item) for item in response.result)
